# This file is part of the faebryk project
# SPDX-License-Identifier: MIT

import logging
from enum import Enum, auto

import faebryk.library._F as F
from faebryk.core.module import Module
from faebryk.libs.library import L
from faebryk.libs.picker.picker import DescriptiveProperties
from faebryk.libs.units import P
from faebryk.libs.util import assert_once

logger = logging.getLogger(__name__)


class USB2514B(Module):
    class ConfigurableUSB(Module):
        """
        USB port wrapper with configuration pins and power enable pin.
        """

        usb: F.USB2_0_IF.Data
        usb_power_enable: F.ElectricLogic
        over_current_sense: F.ElectricLogic

        # configuration interfaces
        battery_charging_enable: F.ElectricLogic
        usb_port_disable_p: F.ElectricLogic
        usb_port_disable_n: F.ElectricLogic

        @assert_once
        def configure_usb_port(
            self,
            owner: Module,
            enable_usb: bool = True,
            enable_battery_charging: bool = True,
        ):
            """
            Configure the specified USB port.
            """
            # enable/disable usb port
            if not enable_usb:
                self.usb_port_disable_p.set_weak(on=True, owner=owner)
                self.usb_port_disable_n.set_weak(on=True, owner=owner)

            # enable/disable battery charging
            if not enable_battery_charging:
                self.battery_charging_enable.set_weak(on=False, owner=owner)

        def __preinit__(self):
            F.ElectricLogic.connect_all_module_references(self)
            self.usb_port_disable_p.line.connect(self.usb.p.line)
            self.usb_port_disable_n.line.connect(self.usb.n.line)
            self.usb_power_enable.connect(self.battery_charging_enable)

    class ConfigurationSource(Enum):
        DEFAULT = auto()
        """
        - Strap options enabled
        - Self-powered operation enabled
        - Individual power switching
        - Individual over-current sensing
        """
        BUS_POWERED = auto()
        """
        Default configuration with the following overrides:
        - Bus-powered operation
        """
        SMBUS = auto()
        """"
        The hub is configured externally over SMBus (as an SMBus slave device):
        - Strap options disabled
        - All registers configured over SMBus
        """
        EEPROM = auto()
        """
        The hub is configured over 2-wire I2C EEPROM:
        - Strap options disabled
        - All registers configured by I2C EEPROM
        """

    @assert_once
    def set_configuration_source(
        self,
        owner: Module,
        configuration_source: ConfigurationSource = ConfigurationSource.DEFAULT,
    ):
        """
        Set the source of configuration settings for the USB2514B.
        """
        if configuration_source == USB2514B.ConfigurationSource.DEFAULT:
            self.configuration_source_input[0].pulled.pull(up=False, owner=owner)
            self.configuration_source_input[1].pulled.pull(up=False, owner=owner)
        elif configuration_source == USB2514B.ConfigurationSource.BUS_POWERED:
            self.configuration_source_input[0].pulled.pull(up=False, owner=owner)
            self.configuration_source_input[1].pulled.pull(up=True, owner=owner)
        elif configuration_source == USB2514B.ConfigurationSource.SMBUS:
            self.configuration_source_input[0].pulled.pull(up=True, owner=owner)
            self.configuration_source_input[1].pulled.pull(up=False, owner=owner)
        elif configuration_source == USB2514B.ConfigurationSource.EEPROM:
            self.configuration_source_input[0].pulled.pull(up=True, owner=owner)
            self.configuration_source_input[1].pulled.pull(up=True, owner=owner)

    class NonRemovablePortConfiguration(Enum):
        ALL_PORTS_REMOVABLE = auto()
        PORT_0_NOT_REMOVABLE = auto()
        PORT_0_1_NOT_REMOVABLE = auto()
        PORT_0_1_2_NOT_REMOVABLE = auto()

    @assert_once
    def set_non_removable_ports(
        self,
        owner: Module,
        non_removable_port_configuration: NonRemovablePortConfiguration,
    ):
        """
        Set the non-removable port configuration of the USB2514.
        """
        if (
            non_removable_port_configuration
            == USB2514B.NonRemovablePortConfiguration.ALL_PORTS_REMOVABLE
        ):
            self.usb_removability_configuration_intput[0].set_weak(
                on=False, owner=owner
            )
            self.usb_removability_configuration_intput[1].set_weak(
                on=False, owner=owner
            )
        elif (
            non_removable_port_configuration
            == USB2514B.NonRemovablePortConfiguration.PORT_0_NOT_REMOVABLE
        ):
            self.usb_removability_configuration_intput[0].set_weak(on=True, owner=owner)
            self.usb_removability_configuration_intput[1].set_weak(
                on=False, owner=owner
            )
        elif (
            non_removable_port_configuration
            == USB2514B.NonRemovablePortConfiguration.PORT_0_1_NOT_REMOVABLE
        ):
            self.usb_removability_configuration_intput[0].set_weak(
                on=False, owner=owner
            )
            self.usb_removability_configuration_intput[1].set_weak(on=True, owner=owner)
        elif (
            non_removable_port_configuration
            == USB2514B.NonRemovablePortConfiguration.PORT_0_1_2_NOT_REMOVABLE
        ):
            self.usb_removability_configuration_intput[0].set_weak(on=True, owner=owner)
            self.usb_removability_configuration_intput[1].set_weak(on=True, owner=owner)

    # ----------------------------------------
    #     modules, interfaces, parameters
    # ----------------------------------------
    power_3v3_regulator: F.ElectricPower
    power_3v3_analog: F.ElectricPower
    power_pll: F.ElectricPower
    power_core: F.ElectricPower
    power_io: F.ElectricPower

    usb_upstream: F.USB2_0_IF.Data
    configurable_downstream_usb = L.list_field(4, ConfigurableUSB)

    xtal_if: F.XtalIF
    external_clock_input: F.ElectricLogic

    usb_bias_resistor_input: F.ElectricSignal
    vbus_detect: F.ElectricSignal

    test: F.Electrical
    reset: F.ElectricLogic
    local_power_detection: F.ElectricSignal

    usb_removability_configuration_intput = L.list_field(2, F.ElectricLogic)
    configuration_source_input = L.list_field(2, F.ElectricLogic)

    suspense_indicator: F.ElectricLogic
    high_speed_upstream_indicator: F.ElectricLogic

    i2c: F.I2C

    # ----------------------------------------
    #                 traits
    # ----------------------------------------
    designator_prefix = L.f_field(F.has_designator_prefix)(
        F.has_designator_prefix.Prefix.U
    )
    descriptive_properties = L.f_field(F.has_descriptive_properties_defined)(
        {
            DescriptiveProperties.manufacturer: "Microchip Tech",
            DescriptiveProperties.partno: "USB2514B-AEZC-TR",
        }
    )
    datasheet = L.f_field(F.has_datasheet_defined)(
        "https://ww1.microchip.com/downloads/aemDocuments/documents/UNG/ProductDocuments/DataSheets/USB251xB-xBi-Data-Sheet-DS00001692.pdf"
    )

    @L.rt_field
    def can_attach_to_footprint(self):
        return F.can_attach_to_footprint_via_pinmap(
            {
                "1": self.configurable_downstream_usb[0].usb.n.line,
                "2": self.configurable_downstream_usb[0].usb.p.line,
                "3": self.configurable_downstream_usb[1].usb.n.line,
                "4": self.configurable_downstream_usb[1].usb.p.line,
                "5": self.power_3v3_analog.hv,
                "6": self.configurable_downstream_usb[2].usb.n.line,
                "7": self.configurable_downstream_usb[2].usb.p.line,
                "8": self.configurable_downstream_usb[3].usb.n.line,
                "9": self.configurable_downstream_usb[3].usb.p.line,
                "10": self.power_3v3_analog.hv,
                "11": self.test,
                "12": self.configurable_downstream_usb[0].battery_charging_enable.line,
                "13": self.configurable_downstream_usb[0].over_current_sense.line,
                "14": self.power_core.hv,
                "15": self.power_3v3_regulator.hv,
                "16": self.configurable_downstream_usb[1].battery_charging_enable.line,
                "17": self.configurable_downstream_usb[1].over_current_sense.line,
                "18": self.configurable_downstream_usb[2].battery_charging_enable.line,
                "19": self.configurable_downstream_usb[2].over_current_sense.line,
                "20": self.configurable_downstream_usb[3].battery_charging_enable.line,
                "21": self.configurable_downstream_usb[3].over_current_sense.line,
                "22": self.usb_removability_configuration_intput[1].line,
                "23": self.power_io.hv,
                "24": self.configuration_source_input[0].line,
                "25": self.configuration_source_input[1].line,
                "26": self.reset.line,
                "27": self.vbus_detect.line,
                "28": self.usb_removability_configuration_intput[0].line,
                "29": self.power_3v3_analog.hv,
                "30": self.usb_upstream.n.line,
                "31": self.usb_upstream.p.line,
                "32": self.xtal_if.xout,
                "33": self.xtal_if.xin,
                "34": self.power_pll.hv,
                "35": self.usb_bias_resistor_input.line,
                "36": self.power_3v3_analog.hv,
                "37": self.power_3v3_analog.lv,
            }
        )

    @L.rt_field
    def pin_association_heuristic(self):
        return F.has_pin_association_heuristic_lookup_table(
            mapping={
                self.power_core.hv: ["CRFILT"],
                self.power_core.lv: ["EP"],
                self.configuration_source_input[1].line: ["HS_IND/CFG_SEL1"],
                self.configurable_downstream_usb[0].over_current_sense.line: ["OCS_N1"],
                self.configurable_downstream_usb[1].over_current_sense.line: ["OCS_N2"],
                self.configurable_downstream_usb[2].over_current_sense.line: ["OCS_N3"],
                self.configurable_downstream_usb[3].over_current_sense.line: ["OCS_N4"],
                self.power_pll.hv: ["PLLFILT"],
                self.configurable_downstream_usb[0].battery_charging_enable.line: [
                    "PRTPWR1/BC_EN1"
                ],
                self.configurable_downstream_usb[1].battery_charging_enable.line: [
                    "PRTPWR2/BC_EN2"
                ],
                self.configurable_downstream_usb[2].battery_charging_enable.line: [
                    "PRTPWR3/BC_EN3"
                ],
                self.configurable_downstream_usb[3].battery_charging_enable.line: [
                    "PRTPWR4/BC_EN4"
                ],
                self.usb_bias_resistor_input.line: ["RBIAS"],
                self.reset.line: ["RESET_N"],
                self.configuration_source_input[0].line: ["SCL/SMBCLK/CFG_SEL0"],
                self.usb_removability_configuration_intput[1].line: [
                    "SDA/SMBDATA/NON_REM1"
                ],
                self.usb_removability_configuration_intput[0].line: [
                    "SUSP_IND/LOCAL_PWR/NON_REM0"
                ],
                self.test: ["TEST"],
                self.configurable_downstream_usb[0].usb.n.line: [
                    "USBDM_DN1/PRT_DIS_M1"
                ],
                self.configurable_downstream_usb[1].usb.n.line: [
                    "USBDM_DN2/PRT_DIS_M2"
                ],
                self.configurable_downstream_usb[2].usb.n.line: [
                    "USBDM_DN3/PRT_DOS_M3"
                ],
                self.configurable_downstream_usb[3].usb.n.line: [
                    "USBDM_DN4/PRT_DIS_M4"
                ],
                self.usb_upstream.p.line: ["USBDM_UP"],
                self.configurable_downstream_usb[0].usb.p.line: [
                    "USBDP_DN1/PRT_DIS_P1"
                ],
                self.configurable_downstream_usb[1].usb.p.line: [
                    "USBDP_DN2/PRT_DIS_P2"
                ],
                self.configurable_downstream_usb[2].usb.p.line: [
                    "USBDP_DN3/PRT_DIS_P3"
                ],
                self.configurable_downstream_usb[3].usb.p.line: [
                    "USBDP_DN4/PRT_DIS_P4"
                ],
                self.usb_upstream.p.line: ["USBDP_UP"],
                self.vbus_detect.line: ["VBUS_DET"],
                self.power_3v3_regulator.hv: ["VDD33"],
                self.power_3v3_analog.hv: ["VDDA33"],
                self.xtal_if.xin: ["XTALIN/CLKIN"],
                self.xtal_if.xout: ["XTALOUT"],
            },
            accept_prefix=False,
            case_sensitive=False,
        )

    def __preinit__(self):
        # ----------------------------------------
        #              connections
        # ----------------------------------------
        self.configuration_source_input[0].connect(self.i2c.scl)
        self.configuration_source_input[1].connect(self.high_speed_upstream_indicator)
        self.usb_removability_configuration_intput[0].line.connect(
            self.suspense_indicator.line,
            self.local_power_detection.line,
        )
        self.usb_removability_configuration_intput[1].connect(self.i2c.sda)
        self.test.connect(self.power_core.lv)

        F.ElectricLogic.connect_all_module_references(self, gnd_only=True)
        F.ElectricLogic.connect_all_module_references(
            self,
            exclude={
                self.power_pll,
                self.power_core,
                # self.power_io,
                self.vbus_detect,
                self.local_power_detection,
                self.power_3v3_regulator,
                self.power_3v3_analog,
            },
        )

        # ----------------------------------------
        #              parametrization
        # ----------------------------------------
        self.power_pll.voltage.constrain_subset(
            L.Range.from_center_rel(1.8 * P.V, 0.05)
        )  # datasheet does not specify a voltage range
        self.power_core.voltage.constrain_subset(
            L.Range.from_center_rel(1.8 * P.V, 0.05)
        )  # datasheet does not specify a voltage range
        self.power_3v3_regulator.voltage.constrain_subset(
            L.Range.from_center(3.3 * P.V, 0.3 * P.V)
        )
        self.power_3v3_analog.voltage.constrain_subset(
            L.Range.from_center(3.3 * P.V, 0.3 * P.V)
        )
        self.power_io.voltage.constrain_subset(
            L.Range.from_center(3.3 * P.V, 0.3 * P.V)
        )



# This file is part of the faebryk project
# SPDX-License-Identifier: MIT

import logging
from pathlib import Path

from faebryk.libs.exceptions import UserResourceException, accumulate
from faebryk.libs.kicad.fileformats import C_kicad_footprint_file
from faebryk.libs.kicad.fileformats_common import C_kicad_footprint_file_header
from faebryk.libs.kicad.fileformats_v5 import C_kicad_footprint_file_v5
from faebryk.libs.kicad.fileformats_v6 import C_kicad_footprint_file_v6
from faebryk.libs.sexp.dataclass_sexp import DecodeError, loads
from faebryk.libs.util import try_relative_to

logger = logging.getLogger(__name__)


def kicad_footprint_file(path: Path) -> C_kicad_footprint_file:
    acc = accumulate(DecodeError, group_message="No decoders succeeded")
    if path.read_text().startswith("(module"):
        with acc.collect():
            return loads(path, C_kicad_footprint_file_v5).convert_to_new()
    else:
        header = loads(path, C_kicad_footprint_file_header)
        version = header.footprint.version
        if version < 20240101:
            with acc.collect():
                return loads(path, C_kicad_footprint_file_v6).convert_to_new()
        else:
            with acc.collect():
                return loads(path, C_kicad_footprint_file)

            with acc.collect():
                return loads(path, C_kicad_footprint_file, ignore_assertions=True)

    # Nothing succeeded in loading the file
    raise UserResourceException(
        f"Footprint {try_relative_to(path.resolve())} is not"
        " a valid KiCad footprint file",
        markdown=False,
    ) from acc.get_exception()



# This file is part of the faebryk project
# SPDX-License-Identifier: MIT

"""
This file contains a faebryk sample.
"""

import logging

import faebryk.library._F as F
from faebryk.core.module import Module
from faebryk.exporters.pcb.layout.extrude import LayoutExtrude
from faebryk.exporters.pcb.layout.typehierarchy import LayoutTypeHierarchy
from faebryk.exporters.pcb.routing.util import Path
from faebryk.libs.library import L
from faebryk.libs.units import P
from faebryk.libs.util import times

logger = logging.getLogger(__name__)


class SubArray(Module):
    unnamed = L.list_field(2, F.Electrical)
    resistors = L.list_field(2, F.Resistor)

    def __init__(self, extrude_y: float):
        super().__init__()
        self._extrude_y = extrude_y

    def __preinit__(self):
        for resistor in self.resistors:
            resistor.resistance.constrain_subset(
                L.Range.from_center_rel(1000 * P.ohm, 0.05)
            )
            resistor.unnamed[0].connect(self.unnamed[0])
            resistor.unnamed[1].connect(self.unnamed[1])

    @L.rt_field
    def pcb_layout(self):
        return F.has_pcb_layout_defined(
            LayoutTypeHierarchy(
                layouts=[
                    LayoutTypeHierarchy.Level(
                        mod_type=F.Resistor,
                        layout=LayoutExtrude((0, self._extrude_y)),
                    ),
                ]
            )
        )

    pcb_routing_strategy = L.f_field(F.has_pcb_routing_strategy_greedy_direct_line)(
        F.has_pcb_routing_strategy_greedy_direct_line.Topology.DIRECT
    )

    @L.rt_field
    def pcb_routing_stategy_manual(self):
        return F.has_pcb_routing_strategy_manual(
            [
                (
                    [r.unnamed[1] for r in self.resistors],
                    Path(
                        [
                            Path.Track(
                                0.1,
                                "F.Cu",
                                [
                                    (0, 0),
                                    (2.5, 0),
                                    (2.5, self._extrude_y),
                                    (0, self._extrude_y),
                                ],
                            ),
                        ]
                    ),
                ),
                (
                    [r.unnamed[0] for r in self.resistors],
                    Path(
                        [
                            Path.Track(
                                0.1,
                                "F.Cu",
                                [
                                    (0, 0),
                                    (-2.5, 0),
                                    (-2.5, self._extrude_y),
                                    (0, self._extrude_y),
                                ],
                            ),
                        ]
                    ),
                ),
            ]
        )


class ResistorArray(Module):
    unnamed = L.list_field(2, F.Electrical)

    @L.rt_field
    def resistors(self):
        return times(self._count, lambda: SubArray(self._extrude_y[1]))

    def __init__(self, count: int, extrude_y: tuple[float, float]):
        super().__init__()

        self._count = count
        self._extrude_y = extrude_y

    def __preinit__(self):
        for resistor in self.resistors:
            resistor.unnamed[0].connect(self.unnamed[0])
            resistor.unnamed[1].connect(self.unnamed[1])

    @L.rt_field
    def pcb_layout(self):
        return F.has_pcb_layout_defined(
            LayoutTypeHierarchy(
                layouts=[
                    LayoutTypeHierarchy.Level(
                        mod_type=SubArray,
                        layout=LayoutExtrude((0, self._extrude_y[0])),
                    ),
                ]
            )
        )

    pcb_routing_strategy = L.f_field(F.has_pcb_routing_strategy_greedy_direct_line)(
        F.has_pcb_routing_strategy_greedy_direct_line.Topology.DIRECT
    )


class App(Module):
    @L.rt_field
    def arrays(self):
        return times(2, lambda: ResistorArray(self._count, self._extrude_y))

    def __init__(
        self, count: int = 2, extrude_y: tuple[float, float] = (15, 5)
    ) -> None:
        super().__init__()
        self._count = count
        self._extrude_y = extrude_y

    def __preinit__(self):
        self.arrays[0].unnamed[1].connect(self.arrays[1].unnamed[0])

    pcb_layout = L.f_field(F.has_pcb_layout_defined)(
        LayoutTypeHierarchy(
            layouts=[
                LayoutTypeHierarchy.Level(
                    mod_type=ResistorArray,
                    layout=LayoutExtrude((10, 0)),
                ),
            ]
        )
    )

    pcb_position = L.f_field(F.has_pcb_position_defined)(
        F.has_pcb_position.Point((20, 20, 0, F.has_pcb_position.layer_type.TOP_LAYER))
    )

    pcb_routing_strategy = L.f_field(F.has_pcb_routing_strategy_greedy_direct_line)(
        F.has_pcb_routing_strategy_greedy_direct_line.Topology.STAR
    )



# This file is part of the faebryk project
# SPDX-License-Identifier: MIT


from faebryk.core.graph import Graph
from faebryk.core.module import Module
from faebryk.libs.app.erc import simple_erc


def run_checks(app: Module, G: Graph):
    simple_erc(G)



# This file is part of the faebryk project
# SPDX-License-Identifier: MIT

from dataclasses import fields
from pathlib import Path

import pytest
from dataclasses_json import CatchAll
from dataclasses_json.utils import CatchAllVar

from atopile.packages import KNOWN_PACKAGES_TO_FOOTPRINT
from faebryk.libs.kicad.fileformats import C_kicad_footprint_file
from faebryk.libs.sexp.dataclass_sexp import dataclass_dfs


def load_footprint(package_path: Path) -> C_kicad_footprint_file:
    """Load a footprint file and return the parsed dataclass."""
    return C_kicad_footprint_file.loads(package_path.read_text())


@pytest.mark.parametrize("package_path", KNOWN_PACKAGES_TO_FOOTPRINT.values())
def test_no_unknown_fields(package_path: Path):
    """Test that no fields in the footprint are marked as unknown.

    This test ensures that all fields in the footprint file can be properly parsed
    into our dataclass structure without falling back to CatchAll fields, which would
    indicate a mismatch between our schema and the file format.
    """
    footprint = load_footprint(package_path)

    # Walk through all dataclass instances in the footprint
    for obj, path, name_path in dataclass_dfs(footprint):
        # dataclass_dfs returns all objects in the tree, including non-dataclass ones
        # like strings and numbers. We only want to check dataclass instances, which
        # have the special __dataclass_fields__ attribute.
        if not hasattr(obj, "__dataclass_fields__"):
            continue

        # Find any CatchAll fields in this dataclass. CatchAll fields are used by
        # dataclasses-json to store any JSON fields that don't match our schema.
        catch_all_fields = [f for f in fields(obj) if f.type in (CatchAll, CatchAllVar)]

        # For each CatchAll field, verify it's None (no unknown fields). A non-None
        # value would indicate that the file contained data we weren't expecting
        # and couldn't properly parse into our schema.
        for field in catch_all_fields:
            value = getattr(obj, field.name)
            assert value is None, (
                f"Found non-None CatchAll field '{field.name}' at {'.'.join(name_path)}"
                f" with value: {value}"
            )



# This file is part of the faebryk project
# SPDX-License-Identifier: MIT

import logging

import faebryk.library._F as F  # noqa: F401
from faebryk.core.module import Module
from faebryk.libs.library import L  # noqa: F401
from faebryk.libs.units import P  # noqa: F401

logger = logging.getLogger(__name__)


class CH344Q_ReferenceDesign(Module):
    """
    Minimal implementation of the CH344Q quad UART to USB bridge
    """

    # ----------------------------------------
    #     modules, interfaces, parameters
    # ----------------------------------------
    usb: F.USB2_0
    usb_uart_converter: F.CH344Q
    oscillator: F.Crystal_Oscillator
    ldo: F.LDO
    led_rx = L.f_field(F.LEDIndicator)(use_mosfet=False)
    led_tx = L.f_field(F.LEDIndicator)(use_mosfet=False)
    led_act = L.f_field(F.LEDIndicator)(use_mosfet=False)
    power_led: F.PoweredLED
    reset_lowpass: F.FilterElectricalRC

    # ----------------------------------------
    #                 traits
    # ----------------------------------------
    @L.rt_field
    def pcb_layout(self):
        from faebryk.exporters.pcb.layout.absolute import LayoutAbsolute
        from faebryk.exporters.pcb.layout.extrude import LayoutExtrude
        from faebryk.exporters.pcb.layout.typehierarchy import LayoutTypeHierarchy

        Point = F.has_pcb_position.Point
        L = F.has_pcb_position.layer_type
        LVL = LayoutTypeHierarchy.Level

        return F.has_pcb_layout_defined(
            layout=LayoutTypeHierarchy(
                layouts=[
                    LVL(
                        mod_type=F.CH344Q,
                        layout=LayoutAbsolute(Point((0, 0, 0, L.NONE))),
                    ),
                    LVL(
                        mod_type=F.Crystal_Oscillator,
                        layout=LayoutAbsolute(Point((-1, 10.75, 180, L.NONE))),
                    ),
                    LVL(
                        mod_type=F.LDO,
                        layout=LayoutAbsolute(Point((7.5, -9.25, 270, L.NONE))),
                    ),
                    LVL(
                        mod_type=F.LEDIndicator,
                        layout=LayoutExtrude(
                            base=Point((8, 9.5, 0, L.NONE)),
                            vector=(-1.75, 0, 90),
                            reverse_order=True,
                        ),
                    ),
                    LVL(
                        mod_type=F.PoweredLED,
                        layout=LayoutAbsolute(Point((-6.5, 9.5, 270, L.NONE))),
                    ),
                    LVL(
                        mod_type=F.FilterElectricalRC,
                        layout=LayoutAbsolute(Point((0, -8, 0, L.NONE))),
                    ),
                ]
            )
        )

    def __preinit__(self):
        # ------------------------------------
        #             aliases
        # ------------------------------------
        pwr_3v3 = self.usb_uart_converter.power
        # ------------------------------------
        #           connections
        # ------------------------------------
        self.usb_uart_converter.power.decoupled.decouple(owner=self).specialize(
            F.MultiCapacitor(4)
        ).set_equal_capacitance_each(L.Range.from_center_rel(100 * P.nF, 0.05))
        self.usb.usb_if.buspower.connect_via(self.ldo, pwr_3v3)

        self.usb.usb_if.d.connect(self.usb_uart_converter.usb)

        self.usb_uart_converter.act.connect(self.led_act.logic_in)
        self.usb_uart_converter.indicator_rx.connect(self.led_rx.logic_in)
        self.usb_uart_converter.indicator_tx.connect(self.led_tx.logic_in)
        pwr_3v3.connect(self.power_led.power)

        self.usb_uart_converter.osc[1].connect(self.oscillator.xtal_if.xin)
        self.usb_uart_converter.osc[0].connect(self.oscillator.xtal_if.xout)
        self.oscillator.xtal_if.gnd.connect(pwr_3v3.lv)

        self.reset_lowpass.out.line.connect(self.usb_uart_converter.reset.line)
        self.reset_lowpass.in_.line.connect(self.usb_uart_converter.reset.reference.hv)
        self.reset_lowpass.in_.reference.connect(
            self.usb_uart_converter.reset.reference
        )
        # TODO: already done by lowpass filter
        # self.usb_uart_converter.reset.pulled.pull(up=True)

        self.usb_uart_converter.test.set_weak(on=False, owner=self)

        self.ldo.enable_output()

        # ------------------------------------
        #          parametrization
        # ------------------------------------
        self.usb_uart_converter.enable_status_or_modem_signals(self)

        self.oscillator.crystal.frequency.constrain_subset(
            L.Range.from_center_rel(8 * P.MHz, 0.001)
        )
        self.oscillator.crystal.frequency_tolerance.constrain_le(40 * P.ppm)
        self.oscillator.crystal.load_capacitance.constrain_subset(
            L.Range.from_center(8 * P.pF, 10 * P.pF)
        )  # TODO: should be property of crystal when picked
        self.oscillator.current_limiting_resistor.resistance.constrain_subset(0 * P.ohm)

        self.ldo.power_in.decoupled.decouple(owner=self).capacitance.constrain_subset(
            L.Range.from_center_rel(100 * P.nF, 0.1)
        )
        self.ldo.power_out.decoupled.decouple(owner=self).capacitance.constrain_subset(
            L.Range.from_center_rel(100 * P.nF, 0.1)
        )

        # self.usb.usb_if.buspower.max_current.constrain_subset(
        #    L.Range.from_center_rel(500 * P.mA, 0.1)
        # )

        self.ldo.output_current.constrain_ge(500 * P.mA)
        self.ldo.output_voltage.constrain_subset(
            L.Range.from_center_rel(3.3 * P.V, 0.05)
        )

        # reset lowpass
        self.reset_lowpass.cutoff_frequency.constrain_subset(
            L.Range.from_center_rel(100 * P.Hz, 0.1)
        )
        self.reset_lowpass.resistor.resistance.constrain_subset(
            L.Range.from_center_rel(4.7 * P.kohm, 0.05)
        )



# Original code developed by : Robert Einhorn
# Modified for us in the atopile parser

import sys
from typing import TYPE_CHECKING, Any, Protocol, TextIO

from antlr4 import InputStream, Lexer, Token
from antlr4.Token import CommonToken

if TYPE_CHECKING:

    class GeneratedAtoLexerProtocol(Protocol):
        @property
        def OPEN_PAREN(self) -> int: ...

        @property
        def OPEN_BRACK(self) -> int: ...

        @property
        def OPEN_BRACE(self) -> int: ...

        @property
        def CLOSE_PAREN(self) -> int: ...

        @property
        def CLOSE_BRACK(self) -> int: ...

        @property
        def CLOSE_BRACE(self) -> int: ...

        @property
        def NEWLINE(self) -> int: ...

        @property
        def INDENT(self) -> int: ...

        @property
        def DEDENT(self) -> int: ...

        @property
        def WS(self) -> int: ...

        @property
        def COMMENT(self) -> int: ...

        @property
        def ERRORTOKEN(self) -> int: ...

        @property
        def symbolicNames(self) -> list[str]: ...

        @property
        def _modeStack(self) -> list[str]: ...

        @property
        def _mode(self) -> list[str]: ...
else:
    GeneratedAtoLexerProtocol = Any


class AtoLexerBase(Lexer, GeneratedAtoLexerProtocol):
    def __init__(self, input: InputStream, output: TextIO = sys.stdout):
        super().__init__(input, output)

        # A stack that keeps track of the indentation lengths
        self.__indent_length_stack: list[int]

        # A list where tokens are waiting to be loaded into the token stream
        self.__pending_tokens: list[CommonToken]

        # last pending token types
        self.__previous_pending_token_type: int
        self.__last_pending_token_type_from_default_channel: int

        # The amount of opened parentheses, square brackets or curly braces
        self.__opened: int
        # The amount of opened parentheses and square brackets in the current lexer mode
        self.__paren_or_bracket_opened_stack: list[int]

        self.__was_space_indentation: bool
        self.__was_tab_indentation: bool
        self.__was_indentation_mixed_with_spaces_and_tabs: bool

        self.__cur_token: CommonToken | None  # current (under processing) token
        self.__ffg_token: CommonToken | None  # following (look ahead) token

        self.__INVALID_LENGTH: int = -1
        self.__ERR_TXT: str = " ERROR: "

        self.__init()

    def nextToken(self) -> CommonToken:  # reading the input stream until a return EOF
        self.__check_next_token()
        return self.__pending_tokens.pop(0)  # add the queued token to the token stream

    def reset(self) -> None:
        self.__init()
        super().reset()

    def __init(self) -> None:
        self.__indent_length_stack = []
        self.__pending_tokens = []
        self.__previous_pending_token_type = 0
        self.__last_pending_token_type_from_default_channel = 0
        self.__opened = 0
        self.__paren_or_bracket_opened_stack = []
        self.__was_space_indentation = False
        self.__was_tab_indentation = False
        self.__was_indentation_mixed_with_spaces_and_tabs = False
        self.__cur_token = None
        self.__ffg_token = None

    def __check_next_token(self) -> None:
        if self.__previous_pending_token_type != Token.EOF:
            self.__set_current_and_following_tokens()
            if len(self.__indent_length_stack) == 0:  # We're at the first token
                self.__handle_start_of_input()

            assert self.__cur_token is not None
            match self.__cur_token.type:
                case self.OPEN_PAREN | self.OPEN_BRACK | self.OPEN_BRACE:
                    self.__opened += 1
                    self.__add_pending_token(self.__cur_token)
                case self.CLOSE_PAREN | self.CLOSE_BRACK | self.CLOSE_BRACE:
                    self.__opened -= 1
                    self.__add_pending_token(self.__cur_token)
                case self.NEWLINE:
                    self.__handle_NEWLINE_token()
                case self.ERRORTOKEN:
                    self.__report_lexer_error(
                        "token recognition error at: '" + self.__cur_token.text + "'"
                    )
                    self.__add_pending_token(self.__cur_token)
                case Token.EOF:
                    self.__handle_EOF_token()
                case _:
                    self.__add_pending_token(self.__cur_token)

    def __set_current_and_following_tokens(self) -> None:
        self.__cur_token = (
            super().nextToken() if self.__ffg_token is None else self.__ffg_token
        )

        self.__handle_lexer_modes()

        self.__ffg_token = (
            self.__cur_token
            if self.__cur_token.type == Token.EOF
            else super().nextToken()
        )

    # initialize the _indent_length_stack
    # hide the leading NEWLINE token(s)
    # if exists, find the first statement (not NEWLINE, not EOF token)
    # that comes from the default channel
    # insert a leading INDENT token if necessary
    def __handle_start_of_input(self) -> None:
        assert self.__cur_token is not None

        # initialize the stack with a default 0 indentation length
        self.__indent_length_stack.append(0)  # this will never be popped off
        while self.__cur_token.type != Token.EOF:
            if self.__cur_token.channel == Token.DEFAULT_CHANNEL:
                if self.__cur_token.type == self.NEWLINE:
                    # all the NEWLINE tokens must be ignored before the first statement
                    self.__hide_and_add_pending_token(self.__cur_token)
                else:  # We're at the first statement
                    self.__insert_leading_indent_token()
                    # continue the processing of the
                    # current token with __check_next_token()
                    return
            else:
                self.__add_pending_token(
                    self.__cur_token
                )  # it can be WS, EXPLICIT_LINE_JOINING or COMMENT token
            self.__set_current_and_following_tokens()
        # continue the processing of the EOF token with __check_next_token()

    def __insert_leading_indent_token(self) -> None:

        if self.__previous_pending_token_type == self.WS:
            prev_token: CommonToken = self.__pending_tokens[-1]  # WS token
            if (
                self.__get_indentation_length(prev_token.text) != 0
            ):  # there is an "indentation" before the first statement
                err_msg: str = "first statement indented"
                self.__report_lexer_error(err_msg)
                # insert an INDENT token before the first statement to
                # raise an 'unexpected indent' error later by the parser
                assert self.__cur_token is not None
                self.__create_and_add_pending_token(
                    self.INDENT,
                    Token.DEFAULT_CHANNEL,
                    self.__ERR_TXT + err_msg,
                    self.__cur_token,
                )

    def __handle_NEWLINE_token(self) -> None:
        assert self.__cur_token is not None
        assert self.__ffg_token is not None

        if (
            self.__opened > 0
        ):  # We're in an implicit line joining, ignore the current NEWLINE token
            self.__hide_and_add_pending_token(self.__cur_token)
        else:
            nl_token: CommonToken = (
                self.__cur_token.clone()
            )  # save the current NEWLINE token
            is_looking_ahead: bool = self.__ffg_token.type == self.WS
            if is_looking_ahead:
                self.__set_current_and_following_tokens()  # set the next two tokens

            match self.__ffg_token.type:
                case self.NEWLINE | self.COMMENT:
                    # We're before a blank line or a comment
                    # or type comment or a type ignore comment
                    self.__hide_and_add_pending_token(
                        nl_token
                    )  # ignore the NEWLINE token
                    if is_looking_ahead:
                        self.__add_pending_token(self.__cur_token)  # WS token
                case other:
                    self.__add_pending_token(nl_token)
                    if (
                        is_looking_ahead
                    ):  # We're on a whitespace(s) followed by a statement
                        indentation_length: int = (
                            0
                            if self.__ffg_token.type == Token.EOF
                            else self.__get_indentation_length(self.__cur_token.text)
                        )

                        if indentation_length != self.__INVALID_LENGTH:
                            self.__add_pending_token(self.__cur_token)  # WS token
                            self.__insert_indent_or_dedent_token(
                                indentation_length
                            )  # may insert INDENT token or DEDENT token(s)
                        else:
                            self.__report_error(
                                "inconsistent use of tabs and spaces in indentation"
                            )
                    else:
                        # We're at a newline followed by a statement
                        # (there is no whitespace before the statement)
                        self.__insert_indent_or_dedent_token(
                            0
                        )  # may insert DEDENT token(s)

    def __insert_indent_or_dedent_token(self, indent_length: int) -> None:
        assert self.__indent_length_stack is not None
        assert self.__ffg_token is not None

        prev_indent_length: int = self.__indent_length_stack[-1]  # peek()
        if indent_length > prev_indent_length:
            self.__create_and_add_pending_token(
                self.INDENT, Token.DEFAULT_CHANNEL, None, self.__ffg_token
            )
            self.__indent_length_stack.append(indent_length)
        else:
            while (
                indent_length < prev_indent_length
            ):  # more than 1 DEDENT token may be inserted to the token stream
                self.__indent_length_stack.pop()
                prev_indent_length = self.__indent_length_stack[-1]  # peek()
                if indent_length <= prev_indent_length:
                    self.__create_and_add_pending_token(
                        self.DEDENT, Token.DEFAULT_CHANNEL, None, self.__ffg_token
                    )
                else:
                    self.__report_error("inconsistent dedent")

    def __handle_lexer_modes(self) -> None:
        if self._modeStack:
            assert self.__cur_token is not None
            match self.__cur_token.type:
                case self.OPEN_BRACE:
                    self.pushMode(Lexer.DEFAULT_MODE)
                    self.__paren_or_bracket_opened_stack.append(0)
                case self.OPEN_PAREN | self.OPEN_BRACK:
                    # https://peps.python.org/pep-0498/#lambdas-inside-expressions
                    self.__paren_or_bracket_opened_stack[-1] += (
                        1  # increment the last element (peek() + 1)
                    )
                case self.CLOSE_PAREN | self.CLOSE_BRACK:
                    self.__paren_or_bracket_opened_stack[-1] -= (
                        1  # decrement the last element (peek() - 1)
                    )
                case self.CLOSE_BRACE:
                    match self._mode:
                        case Lexer.DEFAULT_MODE:
                            self.popMode()
                            self.__paren_or_bracket_opened_stack.pop()
                        case _:
                            self.__report_lexer_error(
                                "f-string: single '}' is not allowed"
                            )

    def __insert_trailing_tokens(self) -> None:
        match self.__last_pending_token_type_from_default_channel:
            case self.NEWLINE | self.DEDENT:
                pass  # no trailing NEWLINE token is needed
            case _:
                # insert an extra trailing NEWLINE token that serves
                # as the end of the last statement
                assert self.__ffg_token is not None
                self.__create_and_add_pending_token(
                    self.NEWLINE, Token.DEFAULT_CHANNEL, None, self.__ffg_token
                )  # _ffg_token is EOF
        self.__insert_indent_or_dedent_token(
            0
        )  # Now insert as much trailing DEDENT tokens as needed

    def __handle_EOF_token(self) -> None:
        assert self.__cur_token is not None

        if self.__last_pending_token_type_from_default_channel > 0:
            # there was statement in the input (leading NEWLINE tokens are hidden)
            self.__insert_trailing_tokens()
        self.__add_pending_token(self.__cur_token)

    def __hide_and_add_pending_token(self, ctkn: CommonToken) -> None:
        ctkn.channel = Token.HIDDEN_CHANNEL
        self.__add_pending_token(ctkn)

    def __create_and_add_pending_token(
        self, ttype: int, channel: int, text: str | None, sample_token: CommonToken
    ) -> None:
        ctkn: CommonToken = sample_token.clone()
        ctkn.type = ttype
        ctkn.channel = channel
        ctkn.stop = sample_token.start - 1
        ctkn.text = "<" + self.symbolicNames[ttype] + ">" if text is None else text

        self.__add_pending_token(ctkn)

    def __add_pending_token(self, ctkn: CommonToken) -> None:
        # save the last pending token type because the _pending_tokens
        # list can be empty by the nextToken()
        self.__previous_pending_token_type = ctkn.type
        if ctkn.channel == Token.DEFAULT_CHANNEL:
            self.__last_pending_token_type_from_default_channel = (
                self.__previous_pending_token_type
            )
        self.__pending_tokens.append(ctkn)

    def __get_indentation_length(
        self, indentText: str
    ) -> int:  # the indentText may contain spaces, tabs or form feeds
        TAB_LENGTH: int = 8  # the standard number of spaces to replace a tab to spaces
        length: int = 0
        ch: str
        for ch in indentText:
            match ch:
                case " ":
                    self.__was_space_indentation = True
                    length += 1
                case "\t":
                    self.__was_tab_indentation = True
                    length += TAB_LENGTH - (length % TAB_LENGTH)
                case "\f":  # form feed
                    length = 0

        if self.__was_tab_indentation and self.__was_space_indentation:
            if not self.__was_indentation_mixed_with_spaces_and_tabs:
                self.__was_indentation_mixed_with_spaces_and_tabs = True
                length = self.__INVALID_LENGTH  # only for the first inconsistent indent
        return length

    def __report_lexer_error(self, err_msg: str) -> None:
        assert self.__cur_token is not None

        self.getErrorListenerDispatch().syntaxError(
            self,
            self.__cur_token,
            self.__cur_token.line,
            self.__cur_token.column,
            " LEXER" + self.__ERR_TXT + err_msg,
            None,
        )

    def __report_error(self, err_msg: str) -> None:
        assert self.__ffg_token is not None

        self.__report_lexer_error(err_msg)

        # the ERRORTOKEN will raise an error in the parser
        self.__create_and_add_pending_token(
            self.ERRORTOKEN,
            Token.DEFAULT_CHANNEL,
            self.__ERR_TXT + err_msg,
            self.__ffg_token,
        )



# This file is part of the faebryk project
# SPDX-License-Identifier: MIT

from os import PathLike
from pathlib import Path
from typing import TYPE_CHECKING

import faebryk.library._F as F
from faebryk.libs.library import L
from faebryk.libs.util import times

if TYPE_CHECKING:
    from faebryk.libs.kicad.fileformats import C_kicad_footprint_file


class KicadFootprint(F.Footprint):
    class has_file(F.Footprint.TraitT.decless()):
        """
        Direct reference to a KiCAD footprint file
        """

        def __init__(self, file: PathLike):
            super().__init__()
            self.file = Path(file)

    class has_kicad_identifier(F.Footprint.TraitT.decless()):
        def __init__(self, kicad_identifier: str):
            super().__init__()
            if ":" not in kicad_identifier:
                raise ValueError(
                    'kicad_identifier must be in the format "library:footprint".'
                    " If not, it'll cause downstream problems"
                )
            self.kicad_identifier = kicad_identifier

        def on_obj_set(self):
            # Implicit trait of has_kicad_footprint is added w/ this trait
            # If this changes, create_footprint_library will need to be updated
            fp = self.get_obj(KicadFootprint)
            if not fp.has_trait(F.has_kicad_footprint):
                fp.add(
                    F.has_kicad_manual_footprint(
                        self.kicad_identifier,
                        {fp.pins[i]: pin_name for i, pin_name in fp.pin_names_sorted},
                    )
                )

    def __init__(self, pin_names: list[str]) -> None:
        super().__init__()

        unique_pin_names = sorted(set(pin_names))
        self.pin_names_sorted = list(enumerate(unique_pin_names))

    @L.rt_field
    def pins(self):
        return times(len(self.pin_names_sorted), F.Pad)

    @L.rt_field
    def attach_via_pinmap(self):
        return F.can_attach_via_pinmap_pinlist(
            {pin_name: self.pins[i] for i, pin_name in self.pin_names_sorted}
        )

    @classmethod
    def from_file(cls, fp_file: "C_kicad_footprint_file") -> "KicadFootprint":
        """
        Create based on a footprint file

        Will take the pin names from that file.
        """

        pad_names = [pad.name for pad in fp_file.footprint.pads]
        self = cls(pad_names)
        return self

    @classmethod
    def from_path(cls, fp_path: PathLike) -> "KicadFootprint":
        """
        Create based on a footprint file

        Will take the pin names from that file.
        """
        from faebryk.libs.kicad.fileformats_version import kicad_footprint_file

        self = cls.from_file(kicad_footprint_file(Path(fp_path)))
        self.add(cls.has_file(fp_path))
        return self



# This file is part of the faebryk project
# SPDX-License-Identifier: MIT

import logging

import faebryk.library._F as F
from faebryk.core.module import Module

logger = logging.getLogger(__name__)


class can_be_decoupled_rails(F.can_be_decoupled.impl()):
    def __init__(self, *rails: F.ElectricPower):
        super().__init__()
        assert rails
        self._rails = rails

    def decouple(
        self,
        owner: Module,
        count: int = 1,
    ) -> F.Capacitor:
        caps = [rail.decoupled.decouple(owner, count) for rail in self._rails]
        # TODO
        return caps[0]



# This file is part of the faebryk project
# SPDX-License-Identifier: MIT

import logging
from abc import abstractmethod
from typing import TYPE_CHECKING, Iterable

from faebryk.core.trait import Trait

if TYPE_CHECKING:
    from faebryk.core.module import Module
    from faebryk.core.moduleinterface import ModuleInterface

logger = logging.getLogger(__name__)


class can_specialize(Trait):
    """
    Marks that a module can specialize other modules next to its bases.
    """

    @abstractmethod
    def get_specializable_types(
        self,
    ) -> Iterable[type["Module"] | type["ModuleInterface"]]:
        """
        Returns a list of types that can be specialized by this module (in addition to
        its own type and the types of its bases).
        """
        pass



# This file is part of the faebryk project
# SPDX-License-Identifier: MIT

from abc import abstractmethod
from typing import TYPE_CHECKING

from faebryk.core.trait import Trait

if TYPE_CHECKING:
    from faebryk.exporters.pcb.kicad.transformer import PCB_Transformer
    from faebryk.exporters.pcb.routing.util import Route


# TODO remove transformer from here
class has_pcb_routing_strategy(Trait):
    @abstractmethod
    def calculate(self, transformer: "PCB_Transformer") -> list["Route"]: ...

    def __preinit__(self):
        self.priority = 0.0

    def __repr__(self) -> str:
        return f"{type(self).__name__}(prio={self.priority})"



import fnmatch
import itertools
import logging
import os
from abc import ABC, abstractmethod
from contextlib import _GeneratorContextManager, contextmanager
from contextvars import ContextVar
from enum import Enum
from pathlib import Path
from typing import Any, Callable, Generator, Iterable, Self

from pydantic import (
    AliasChoices,
    BaseModel,
    ConfigDict,
    Field,
    ValidationError,
    ValidationInfo,
    field_serializer,
    field_validator,
    model_validator,
)
from pydantic_settings import (
    BaseSettings,
    PydanticBaseSettingsSource,
    SettingsConfigDict,
    SettingsError,
    YamlConfigSettingsSource,
)
from ruamel.yaml import YAML

from atopile import address, version
from atopile.address import AddressError, AddrStr
from atopile.errors import (
    UserBadParameterError,
    UserException,
    UserFileNotFoundError,
    UserNoProjectException,
    UserNotImplementedError,
)
from atopile.version import DISTRIBUTION_NAME, get_installed_atopile_version
from faebryk.libs.exceptions import UserResourceException, iter_through_errors

logger = logging.getLogger(__name__)
yaml = YAML()


APPLICATION_NAME = "atopile"
ENV_VAR_PREFIX = "ATO_"
PROJECT_CONFIG_FILENAME = "ato.yaml"
GLOBAL_CONFIG_FILENAME = "config.yaml"


def _loc_to_dot_sep(loc: tuple[str | int, ...]) -> str:
    # via https://docs.pydantic.dev/dev/errors/errors/#customize-error-messages
    path = ""
    for i, x in enumerate(loc):
        if isinstance(x, str):
            if i > 0:
                path += "."
            path += x
        elif isinstance(x, int):
            path += f"[{x}]"
        else:
            raise TypeError("Unexpected type")
    return path


def _convert_errors(e: ValidationError) -> list[dict[str, Any]]:
    return [
        {"msg": error["msg"], "loc": _loc_to_dot_sep(error["loc"])}
        for error in e.errors()
    ]


def _try_construct_config[T](
    model: type[T], identifier: str | Path | None = None, **kwargs: Any
) -> T:
    message_prefix = f"`{identifier}`: " if identifier else ""

    try:
        return model(**kwargs)
    except ValidationError as ex:
        excs = ExceptionGroup(
            "Configuration is invalid",
            [
                UserConfigurationError(
                    f"{message_prefix}{error['msg']}: `{error['loc']}`"
                )
                for error in _convert_errors(ex)
            ],
        )
        raise excs from ex
    except SettingsError as ex:
        raise UserConfigurationError(f"Invalid config: {ex}") from ex


class BaseConfigModel(BaseModel):
    model_config = ConfigDict(use_attribute_docstrings=True)


class BuildType(Enum):
    ATO = "ato"
    PYTHON = "python"


class UserConfigurationError(UserException):
    """An error in the config file."""


class UserConfigNotFoundError(UserException):
    """No project config file was found."""


class ConfigFileSettingsSource(YamlConfigSettingsSource, ABC):
    def __init__(self, settings_cls: type[BaseSettings]):
        self.yaml_file_path = self.find_config_file()
        self.yaml_file_encoding = "utf-8"
        self.yaml_data = self._read_files(self.yaml_file_path)

        super(YamlConfigSettingsSource, self).__init__(settings_cls, self.get_data())

    @classmethod
    @abstractmethod
    def find_config_file(cls) -> Path | None: ...

    @abstractmethod
    def get_data(self) -> dict[str, Any]: ...


class GlobalConfigSettingsSource(ConfigFileSettingsSource):
    @classmethod
    def find_config_file(cls) -> Path | None:
        """Find the global config file in the user's home directory."""

        # note deliberate use of ~/.config on all platforms
        # (rather than e.g. platformdirs)
        config_dir = (
            Path(os.environ.get("XDG_CONFIG_HOME", Path.home() / ".config"))
            / APPLICATION_NAME
        )
        config_file = config_dir / GLOBAL_CONFIG_FILENAME
        return config_file if config_file.exists() else None

    def get_data(self) -> dict[str, Any]:
        return self.yaml_data if self.yaml_data else {}


def _find_project_config_file(start: Path) -> Path | None:
    """Search parent directories, up to the root, for a project config file."""
    path = start
    while not (path / PROJECT_CONFIG_FILENAME).exists():
        path = path.parent
        if path == path.parent:
            return None

    return path.resolve().absolute() / PROJECT_CONFIG_FILENAME


class ProjectConfigSettingsSource(ConfigFileSettingsSource):
    @classmethod
    def find_config_file(cls) -> Path | None:
        """
        Find a project config file in the specified directory or any parent directories.
        """
        if _project_dir:
            return _project_dir / PROJECT_CONFIG_FILENAME

        return _find_project_config_file(Path.cwd())

    def get_data(self) -> dict[str, Any]:
        return self.yaml_data or {}


class ProjectPaths(BaseConfigModel):
    """
    Project global paths
    """

    root: Path
    """Project root directory (where the ato.yaml file is located)"""

    src: Path
    """Project source code directory"""

    layout: Path
    """Project layout directory where KiCAD projects are stored and searched for"""

    footprints: Path
    """Project footprints directory"""

    build: Path
    """Build artifact output directory"""

    logs: Path
    """Build logs directory"""

    manifest: Path
    """Build manifest file"""

    component_lib: Path
    """Component library directory for builds"""

    modules: Path
    """Project modules directory (`.ato/modules` from the project root)"""

    def __init__(self, **data: Any):
        data.setdefault("root", _project_dir or Path.cwd())
        data.setdefault("src", data["root"] / "elec" / "src")
        data.setdefault("layout", data["root"] / "elec" / "layout")
        data.setdefault("footprints", data["root"] / "elec" / "footprints")
        data["build"] = Path(data.get("build", data["root"] / "build"))
        data.setdefault("logs", data["build"] / "logs")
        data.setdefault("manifest", data["build"] / "manifest.json")
        data.setdefault("component_lib", data["build"] / "kicad" / "libs")
        data.setdefault("modules", data["root"] / ".ato" / "modules")
        super().__init__(**data)

    @model_validator(mode="after")
    def make_paths_absolute(model: "ProjectPaths") -> "ProjectPaths":
        """Make all paths absolute relative to the project root."""
        if not model.root.is_absolute():
            model.root = model.root.resolve().absolute()

        for field_name, field_value in model:
            if field_name != "root" and isinstance(field_value, Path):
                if not field_value.is_absolute():
                    setattr(model, field_name, model.root / field_value)
                setattr(
                    model, field_name, getattr(model, field_name).resolve().absolute()
                )
        return model

    def ensure(self) -> None:
        self.build.mkdir(parents=True, exist_ok=True)
        self.layout.mkdir(parents=True, exist_ok=True)

    def get_footprint_lib(self, lib_name: str) -> Path:
        return self.component_lib / "footprints" / f"{lib_name}.pretty"


class BuildTargetPaths(BaseConfigModel):
    """
    Build-target specific paths
    """

    layout: Path
    """Build-target layout file"""

    output_base: Path
    """Extension-less filename for build artifacts"""

    netlist: Path
    """Build-target netlist file"""

    fp_lib_table: Path
    """Project footprint library table file"""

    kicad_project: Path
    """Build-target KiCAD project file"""

    def __init__(self, name: str, project_paths: ProjectPaths, **data: Any):
        if layout_data := data.get("layout"):
            data["layout"] = BuildTargetPaths.find_layout(Path(layout_data))
        else:
            data["layout"] = BuildTargetPaths.find_layout(
                project_paths.root / project_paths.layout / name
            )

        if output_base_data := data.get("output_base"):
            data["output_base"] = Path(output_base_data)
        else:
            data["output_base"] = project_paths.build / "builds" / name / name
            data["output_base"].parent.mkdir(parents=True, exist_ok=True)

        data.setdefault("netlist", data["output_base"] / f"{name}.net")
        data.setdefault("fp_lib_table", data["layout"].parent / "fp-lib-table")
        data.setdefault("kicad_project", data["layout"].with_suffix(".kicad_pro"))
        # We deliberately don't set a root for an individual build
        super().__init__(**data)

    @classmethod
    def find_layout(cls, layout_base: Path) -> Path:
        """Find the layout associated with a build."""

        if layout_base.with_suffix(".kicad_pcb").exists():
            return layout_base.with_suffix(".kicad_pcb").resolve().absolute()
        elif layout_base.is_dir():
            layout_candidates = list(
                filter(
                    BuildTargetPaths.match_user_layout, layout_base.glob("*.kicad_pcb")
                )
            )

            if len(layout_candidates) == 1:
                return layout_candidates[0].resolve().absolute()
            else:
                raise UserException(
                    "Layout directories must contain only 1 layout,"
                    f" but {len(layout_candidates)} found in {layout_base}"
                )

        # default location, to create later
        return layout_base.resolve().absolute() / f"{layout_base.name}.kicad_pcb"

    def ensure_layout(self):
        """Return the layout associated with a build."""
        if not self.layout.exists():
            logger.info("Creating new layout at %s", self.layout)
            self.layout.parent.mkdir(parents=True, exist_ok=True)

            # delayed import to improve startup time
            from faebryk.libs.kicad.fileformats import C_kicad_pcb_file

            C_kicad_pcb_file.skeleton(
                generator=DISTRIBUTION_NAME,
                generator_version=str(get_installed_atopile_version()),
            ).dumps(self.layout)
        elif not self.layout.is_file():
            raise UserResourceException(f"Layout is not a file: {self.layout}")

    @classmethod
    def match_user_layout(cls, path: Path) -> bool:
        """Check whether a given filename is a KiCAD autosaved layout."""
        autosave_patterns = [
            "_autosave-*",
            "*-save.kicad_pcb",
        ]

        name = path.name

        for pattern in autosave_patterns:
            if fnmatch.fnmatch(name, pattern):
                return False
        return True


class BuildTargetConfig(BaseConfigModel, validate_assignment=True):
    _project_paths: ProjectPaths

    name: str
    address: str | None = Field(alias="entry")
    """
    Entry point or root node for the build-target.
    Everything that exists within this module will be built as part of this build-target
    """

    targets: list[str] = Field(default=["__default__"])  # TODO: validate
    """A list of targets' names to build after updating the layout"""

    exclude_targets: list[str] = Field(default=[])
    """
    A list of targets' names to exclude.

    Applied after `targets` are specified. This makes it useful to exclude
    targets that aren't relevant to a specific build - especially in CI,
    which typically builds **all** targets.
    """

    fail_on_drcs: bool = Field(default=False)
    dont_solve_equations: bool = Field(default=False)
    keep_designators: bool | None = Field(default=True)
    keep_picked_parts: bool | None = Field(default=None)
    keep_net_names: bool | None = Field(default=None)
    frozen: bool = Field(default=False)
    paths: BuildTargetPaths

    def __init__(self, **data: Any):
        super().__init__(**data)
        self._project_paths = data.get("_project_paths", ProjectPaths())

    @model_validator(mode="before")
    def init_paths(cls, data: dict) -> dict:
        match data.get("paths"):
            case dict() | None:
                data["paths"] = BuildTargetPaths(
                    name=data["name"],
                    project_paths=data["_project_paths"],
                    **data.get("paths", {}),
                )
            case BuildTargetPaths():
                pass
            case _:
                raise ValueError(f"Invalid build paths: {data.get('paths')}")
        return data

    def _set_frozen(self, frozen: bool):
        if self.keep_designators is None:
            self.keep_designators = frozen
        if self.keep_picked_parts is None:
            self.keep_picked_parts = frozen
        if self.keep_net_names is None:
            self.keep_net_names = frozen
        self.frozen = frozen

    @model_validator(mode="after")
    def validate_frozen_after(self) -> Self:
        if self.frozen:
            if not self.keep_designators:
                raise ValueError(
                    "`keep_designators` must be true when `frozen` is true"
                )
            if not self.keep_picked_parts:
                raise ValueError(
                    "`keep_picked_parts` must be true when `frozen` is true"
                )
            if not self.keep_net_names:
                raise ValueError("`keep_net_names` must be true when `frozen` is true")

        return self

    @property
    def build_type(self) -> BuildType:
        """
        Determine build type from the entry.

        **/*.ato:* -> BuildType.ATO
        **/*.py:* -> BuildType.PYTHON
        """
        suffix = self.entry_file_path.suffix

        match suffix:
            case ".ato":
                return BuildType.ATO
            case ".py":
                return BuildType.PYTHON
            case _:
                raise UserException(
                    f"Unknown entry suffix: {suffix} for {self.entry_section}"
                )

    @property
    def entry_file_path(self) -> Path:
        """An absolute path to the entry file."""
        address = AddrStr(self.address)
        return self._project_paths.root / address.file_path

    @property
    def entry_section(self) -> str:
        """The path to the entry module."""
        address = AddrStr(self.address)
        return address.entry_section

    def ensure(self):
        """Ensure this build config is ready to be used"""
        self.paths.ensure_layout()


class Source(BaseConfigModel):
    local: Path | None = None

    @field_serializer("local")
    def serialize_local(self, path: Path | None, _info: Any) -> str | None:
        return str(path) if path else None


class Dependency(BaseConfigModel):
    name: str
    version_spec: str | None = None
    link_broken: bool = False
    path: Path | None = None
    source: Source | None = None

    project_config: "ProjectConfig | None" = Field(
        default_factory=lambda data: ProjectConfig.from_path(data["path"]), exclude=True
    )

    @classmethod
    def from_str(cls, spec_str: str) -> "Dependency":
        for splitter in version.OPERATORS + ("@",):
            if splitter in spec_str:
                try:
                    name, version_spec = spec_str.rsplit(splitter, 1)
                    name = name.strip()
                    version_spec = splitter + version_spec
                except TypeError as ex:
                    raise UserConfigurationError(
                        f"Invalid dependency spec: {spec_str}"
                    ) from ex
                return cls(name=name, version_spec=version_spec)
        return cls(name=spec_str)

    @field_serializer("path")
    def serialize_path(self, path: Path | None, _info: Any) -> str | None:
        return str(path) if path else None


class ServicesConfig(BaseConfigModel):
    class Components(BaseConfigModel):
        url: str = Field(default="https://components.atopileapi.com")
        """Components URL"""

    class Packages(BaseConfigModel):
        url: str = Field(default="https://packages.atopileapi.com")
        """Packages URL"""

    @field_validator("components", mode="before")
    def validate_components(cls, value: Components | str) -> Components:
        # also accepts a string for backwards compatibility
        if isinstance(value, str):
            return cls.Components(url=value)
        return value

    components: Components = Field(default_factory=Components)
    packages: Packages = Field(default_factory=Packages)


class ProjectConfig(BaseConfigModel):
    """Project-level config"""

    ato_version: str = Field(
        validation_alias=AliasChoices("ato-version", "ato_version"),
        serialization_alias="ato-version",
        default=f"{version.get_installed_atopile_version()}",
    )
    """
    The compiler version with which the project was developed.

    This is used by the compiler to ensure the code in this project is
    compatible with the compiler version.
    """

    paths: ProjectPaths = Field(default_factory=ProjectPaths)
    dependencies: list[Dependency] | None = Field(default=None)
    """
    Represents requirements on other projects.

    Typically, you shouldn't modify this directly.

    Instead, use the `ato install` command to install dependencies.
    """

    entry: str | None = Field(default=None)
    builds: dict[str, BuildTargetConfig] = Field(default_factory=dict)
    """A map of all the build targets (/ "builds") in this project."""

    services: ServicesConfig = Field(default_factory=ServicesConfig)
    open_layout_on_build: bool | None = None
    """Automatically open pcbnew when applying netlist"""

    @classmethod
    def from_path(cls, path: Path | None) -> "ProjectConfig | None":
        if path is None:
            return None

        config_file = path / PROJECT_CONFIG_FILENAME

        if not config_file.exists():
            return None

        try:
            file_contents = yaml.load(config_file)
        except FileNotFoundError as e:
            raise UserFileNotFoundError(f"Failed to load project config: {e}") from e
        except Exception as e:
            raise UserConfigurationError(f"Failed to load project config: {e}") from e

        file_contents.setdefault("paths", {}).setdefault("root", path)
        return _try_construct_config(
            ProjectConfig, identifier=config_file, **file_contents
        )

    @field_validator("builds", mode="before")
    def init_builds(
        cls, value: dict[str, dict[str, Any] | BuildTargetConfig], info: ValidationInfo
    ) -> dict[str, Any]:
        for build_name, data in value.items():
            match data:
                case BuildTargetConfig():
                    data.name = build_name
                case dict():
                    data.setdefault("name", build_name)
                    data["_project_paths"] = info.data["paths"]
                case _:
                    raise ValueError(f"Invalid build data: {data}")
        return value

    @field_validator("dependencies", mode="before")
    def add_dependencies(cls, value: list[dict[str, Any]] | None) -> list[Dependency]:
        return [
            Dependency.from_str(dep) if isinstance(dep, str) else Dependency(**dep)
            for dep in (value or [])
        ]

    @model_validator(mode="after")
    def validate_compiler_versions(self) -> Self:
        """
        Check that the compiler version is compatible with the version
        used to build the project.
        """
        dependency_cfgs = (
            (dep.project_config for dep in self.dependencies)
            if self.dependencies is not None
            else ()
        )

        for cltr, cfg in iter_through_errors(itertools.chain([self], dependency_cfgs)):
            if cfg is None:
                continue

            with cltr():
                semver_str = cfg.ato_version
                # FIXME: this is a hack to the moment to get around us breaking
                # the versioning scheme in the ato.yaml files
                for operator in version.OPERATORS:
                    semver_str = semver_str.replace(operator, "")

                built_with_version = version.parse(semver_str)

                if not version.match_compiler_compatability(built_with_version):
                    raise version.VersionMismatchError(
                        f"{cfg.paths.root} ({cfg.ato_version}) can't be"
                        " built with this version of atopile "
                        f"({version.get_installed_atopile_version()})."
                    )
        return self

    @classmethod
    def skeleton(cls, entry: str, paths: ProjectPaths | None):
        """Creates a minimal ProjectConfig"""
        project_paths = paths or ProjectPaths()
        return _try_construct_config(
            ProjectConfig,
            ato_version=f"^{version.get_installed_atopile_version()}",
            paths=project_paths,
            entry=entry,
            builds={
                "default": BuildTargetConfig(
                    _project_paths=project_paths,
                    name="default",
                    entry="",
                    paths=BuildTargetPaths(name="default", project_paths=project_paths),
                )
            },
        )


class ProjectSettings(ProjectConfig, BaseSettings):  # FIXME
    """
    Project-level config, loaded from
    - global config e.g. ~/.config/atopile/config.yaml
    - ato.yaml
    - environment variables e.g. ATO_SERVICES_COMPONENTS_API_URL
    """

    # TOOD: ignore but warn for extra fields
    model_config = BaseConfigModel.model_config | SettingsConfigDict(
        env_prefix=ENV_VAR_PREFIX,
        env_nested_delimiter="_",
        enable_decoding=False,
        use_attribute_docstrings=True,
        extra="forbid",
    )

    @classmethod
    def settings_customise_sources(
        cls,
        settings_cls: type[BaseSettings],
        init_settings: PydanticBaseSettingsSource,
        env_settings: PydanticBaseSettingsSource,
        dotenv_settings: PydanticBaseSettingsSource,
        file_secret_settings: PydanticBaseSettingsSource,
    ) -> tuple[PydanticBaseSettingsSource, ...]:
        return (
            init_settings,
            env_settings,
            ProjectConfigSettingsSource(settings_cls),
            GlobalConfigSettingsSource(settings_cls),
        )


_current_build_cfg: ContextVar[BuildTargetConfig | None] = ContextVar(
    "current_build_cfg", default=None
)


@contextmanager
def _build_context(config: "Config", build_name: str):
    cfg = config.project.builds[build_name]
    cfg.ensure()
    token = _current_build_cfg.set(cfg)
    try:
        yield
    finally:
        _current_build_cfg.reset(token)


class Config:
    _project: ProjectSettings | ProjectConfig | None
    _entry: str | None
    _selected_builds: list[str] | None
    _project_dir: Path | None
    interactive: bool

    def __init__(self) -> None:
        self._project_dir = _project_dir
        self._project = _try_construct_config(ProjectSettings)
        self._entry = None
        self._selected_builds = None
        self.interactive = True

    def __repr__(self) -> str:
        return self._project.__repr__()

    def __rich_repr__(self):
        yield "project", self._project
        yield "entry", self._entry
        yield "selected_builds", self._selected_builds
        yield "project_dir", self._project_dir

    @property
    def project(self) -> ProjectSettings | ProjectConfig:
        if self._project is None:
            # TODO: better message
            raise UserConfigurationError("No project config found")

        return self._project

    @project.setter
    def project(self, value: ProjectSettings | ProjectConfig) -> None:
        self._project = value

    @property
    def project_dir(self) -> Path:
        if not self._project_dir:
            raise ValueError("No project directory found")
        return self._project_dir

    @project_dir.setter
    def project_dir(self, value: os.PathLike) -> None:
        global _project_dir
        _project_dir = Path(value)

        if not (_project_dir / PROJECT_CONFIG_FILENAME).is_file():
            raise UserConfigNotFoundError(
                f"No `{PROJECT_CONFIG_FILENAME}` found in the specified directory"
            )

        self._project_dir = _project_dir
        self._project = _try_construct_config(ProjectSettings)

    def update_project_config(
        self, transformer: Callable[[dict, dict], dict], new_data: dict
    ) -> None:
        """Apply an update to the project config file."""
        yaml = YAML(typ="rt")  # round-trip
        filename = self.project_dir / PROJECT_CONFIG_FILENAME
        temp_filename = filename.with_suffix(".yaml.tmp")

        try:
            with filename.open("r", encoding="utf-8") as file:
                yaml_data: dict = yaml.load(filename) or {}

            yaml_data = transformer(yaml_data, new_data)

            with temp_filename.open("w", encoding="utf-8") as file:
                yaml.dump(yaml_data, file)

            temp_filename.replace(filename)

            self._project = _try_construct_config(ProjectSettings)

        except Exception as e:
            try:
                temp_filename.unlink(missing_ok=True)
            except Exception:
                pass
            raise e

    @property
    def selected_builds(self) -> Iterable[str]:
        return self._selected_builds or self.project.builds.keys()

    @selected_builds.setter
    def selected_builds(self, value: list[str]) -> None:
        self._selected_builds = value

    @property
    def builds(self) -> Generator[_GeneratorContextManager[None], None, None]:
        """Return an iterable of BuildContext objects for each build."""
        return (_build_context(self, name) for name in self.selected_builds)

    @property
    def build(self) -> BuildTargetConfig:
        if current := self._current_build:
            return current
        raise RuntimeError("No build config is currently active")

    @property
    def has_project(self) -> bool:
        try:
            project_dir = self.project_dir
        except ValueError:
            return False
        return (project_dir / PROJECT_CONFIG_FILENAME).exists()

    @property
    def _current_build(self) -> BuildTargetConfig | None:
        return _current_build_cfg.get()

    def _setup_standalone(self, entry: str | None, entry_arg_file_path: Path) -> None:
        if not entry:
            raise UserBadParameterError(
                "You must specify an entry to build with the --standalone option"
            )
        if not entry_arg_file_path.exists():
            raise UserBadParameterError(
                f"The file you have specified does not exist: {entry_arg_file_path}"
            )

        if not entry_arg_file_path.is_file():
            raise UserBadParameterError(
                "The path you're building with the --standalone"
                f" option must be a file {entry_arg_file_path}",
                markdown=False,
            )

        if config.has_project:
            raise UserBadParameterError(
                "Project config must not be present for standalone builds"
            )

        try:
            AddrStr(entry).entry_section
        except AddressError:
            raise UserBadParameterError(
                "You must specify what to build within a file to build with the"
                " --standalone option"
            )

        # don't trigger reload
        self._project_dir = entry_arg_file_path.parent or Path.cwd()
        self._project = ProjectConfig.skeleton(
            entry=entry,
            paths=ProjectPaths(
                root=self._project_dir,
                src=self._project_dir,
                layout=self._project_dir / "standalone",
                footprints=self._project_dir / "standalone",
            ),
        )

    def _check_entry_arg_file_path(
        self, entry: str | None, entry_arg_file_path: Path
    ) -> AddrStr | None:
        entry_addr_override = None

        if entry:
            if entry_arg_file_path.is_file():
                if entry_section := address.get_entry_section(AddrStr(entry)):
                    entry_addr_override = address.from_parts(
                        str(entry_arg_file_path.absolute()),
                        entry_section,
                    )
                else:
                    raise UserBadParameterError(
                        "If an entry of a file is specified, you must specify"
                        " the node within it you want to build.",
                        title="Bad 'entry' parameter",
                    )

            elif entry_arg_file_path.is_dir():
                pass

            elif not entry_arg_file_path.exists():
                raise UserBadParameterError(
                    "The entry you have specified does not exist.",
                    title="Bad 'entry' parameter",
                )
            else:
                raise ValueError(
                    f"Unexpected entry path type {entry_arg_file_path}"
                    " - this should never happen!"
                )

        return entry_addr_override

    def _get_entry_arg_file_path(
        self, entry: str | None
    ) -> tuple[AddrStr | None, Path]:
        # basic the entry address if provided, otherwise leave it as None

        if entry is None:
            entry_arg_file_path = Path.cwd()
        else:
            entry = AddrStr(entry)

            if address.get_file(entry) is None:
                raise UserBadParameterError(
                    f"Invalid entry address {entry} - entry must specify a file.",
                    title="Bad 'entry' parameter",
                )

            entry_arg_file_path = (
                Path(address.get_file(entry)).expanduser().resolve().absolute()
            )

        return entry, entry_arg_file_path

    def apply_options(
        self,
        entry: str | None,
        standalone: bool = False,
        option: Iterable[str] = (),
        target: Iterable[str] = (),
        selected_builds: Iterable[str] = (),
        frozen: bool | None = None,
        **kwargs: Any,
    ) -> None:
        entry, entry_arg_file_path = self._get_entry_arg_file_path(entry)

        if standalone:
            self._setup_standalone(entry, entry_arg_file_path)
        else:
            if config_file_path := _find_project_config_file(entry_arg_file_path):
                self.project_dir = config_file_path.parent
            elif entry is None:
                raise UserNoProjectException()

            else:
                raise UserBadParameterError(
                    f"Specified entry path is not a file or directory: "
                    f"{entry_arg_file_path}",
                    markdown=False,
                )

        self.project.entry = entry

        logger.info("Using project %s", self.project_dir)

        # add custom config overrides
        if option:
            raise UserNotImplementedError(
                "Custom config overrides have been removed in a refactor. "
                "It's planned to re-add them in a future release. "
                "If this is a blocker for you, please raise an issue. "
                "In the meantime, you can use the `ato.yaml` file to set these options."
            )

        # if we set an entry-point, we now need to deal with that
        entry_addr_override = self._check_entry_arg_file_path(
            entry, entry_arg_file_path
        )

        if self.project.paths is not None:
            self.project.paths.ensure()

        if selected_builds:
            self.selected_builds = list(selected_builds)

        for build_name in self.selected_builds:
            build_cfg = self.project.builds[build_name]

            if build_name not in self.project.builds:
                raise UserBadParameterError(
                    f"Build `{build_name}` not found in project config"
                )

            if entry_addr_override is not None:
                build_cfg.address = entry_addr_override
            if target:
                build_cfg.targets = list(target)

            # Attach CLI options passed via kwargs
            for key, value in kwargs.items():
                if value is not None:
                    setattr(build_cfg, key, value)

            if frozen is not None:
                try:
                    build_cfg._set_frozen(frozen)
                except ValidationError as e:
                    # TODO: better error message
                    raise UserBadParameterError(
                        f"Invalid value for `frozen`: {e}",
                        title="Bad 'frozen' parameter",
                    )

    def should_open_layout_on_build(self) -> bool:
        """Returns whether atopile should open the layout after building"""
        # If the project config has an explicit setting, use that
        if self.project is not None and self.project.open_layout_on_build is not None:
            return self.project.open_layout_on_build

        # Otherwise, default to opening the layout if we're only building a
        # single target and we're running interactively
        return (
            (self.project is None or self.project.open_layout_on_build is None)
            and len(list(self.selected_builds)) == 1
            and self.interactive
        )


_project_dir: Path | None = None
config: Config = Config()



# This file is part of the faebryk project
# SPDX-License-Identifier: MIT

import faebryk.library._F as F


class can_attach_via_pinmap_pinlist(F.can_attach_via_pinmap.impl()):
    def __init__(self, pin_list: dict[str, F.Pad]) -> None:
        super().__init__()
        self.pin_list = pin_list

    def attach(self, pinmap: dict[str, F.Electrical | None]):
        for no, intf in pinmap.items():
            if intf is None:
                continue
            assert (
                no in self.pin_list
            ), f"Pin {no} not in pin list: {self.pin_list.keys()}"
            self.pin_list[no].attach(intf)



import logging
import sys
from pathlib import Path
from typing import Iterable

from git import Repo

from faebryk.libs.util import AUTO_RECOMPILE, is_editable_install, run_live

logger = logging.getLogger(__name__)


if is_editable_install() and AUTO_RECOMPILE.get():

    def has_uncommitted_changes(files: Iterable[str | Path]) -> bool:
        """Check if any of the given files have uncommitted changes."""
        try:
            repo = Repo(search_parent_directories=True)
            diff_index = repo.index.diff(None)  # Get uncommitted changes

            # Convert all files to Path objects for consistent comparison
            files = [Path(f).resolve() for f in files]
            repo_root = Path(repo.working_dir)

            # Check if any of the files have changes
            for diff in diff_index:
                diff_path = (repo_root / diff.a_path).resolve()
                if diff_path in files:
                    return True

            return False
        except Exception:
            # If we can't check git status (not a git repo, etc), assume we don't need to recompile # noqa: E501  # pre-existing
            return False

    SAUCY_FILES = [
        "AtoParser.g4",
        "AtoLexer.g4",
        "AtoParserBase.py",
        "AtoLexerBase.py",
    ]
    THIS_DIR = Path(__file__).parent

    if has_uncommitted_changes(THIS_DIR / f for f in SAUCY_FILES):
        logger.warning("Recompiling ANTLR4 grammars due to changes in grammar files")
        bin_dir = Path(sys.executable).parent
        run_live(
            [
                bin_dir / "antlr4",
                "-visitor",
                "-no-listener",
                "-Dlanguage=Python3",
                "AtoLexer.g4",
                "AtoParser.g4",
            ],
            cwd=THIS_DIR,
        )



# This file is part of the faebryk project
# SPDX-License-Identifier: MIT

import logging

import faebryk.library._F as F  # noqa: F401
from faebryk.core.module import Module
from faebryk.libs.library import L  # noqa: F401
from faebryk.libs.units import P  # noqa: F401

logger = logging.getLogger(__name__)


class DE9RS232Connector(Module):
    """
    Standard RS232 bus on DE-9 connector
    """

    # ----------------------------------------
    #     modules, interfaces, parameters
    # ----------------------------------------
    connector: F.DE9Connector
    rs232: F.RS232

    # ----------------------------------------
    #                 traits
    # ----------------------------------------

    @L.rt_field
    def single_electric_reference(self):
        return F.has_single_electric_reference_defined(
            F.ElectricLogic.connect_all_module_references(self)
        )

    def __preinit__(self):
        # ------------------------------------
        #           connections
        # ------------------------------------
        self.rs232.tx.line.connect(self.connector.contact[2])
        self.rs232.rx.line.connect(self.connector.contact[1])
        self.rs232.dtr.line.connect(self.connector.contact[3])
        self.rs232.dcd.line.connect(self.connector.contact[0])
        self.rs232.dsr.line.connect(self.connector.contact[5])
        self.rs232.ri.line.connect(self.connector.contact[8])
        self.rs232.rts.line.connect(self.connector.contact[6])
        self.rs232.cts.line.connect(self.connector.contact[7])

        self.rs232.get_trait(
            F.has_single_electric_reference
        ).get_reference().lv.connect(self.connector.contact[4])

        # ------------------------------------
        #          parametrization
        # ------------------------------------



import faebryk.library._F as F
from faebryk.core.module import Module
from faebryk.core.moduleinterface import ModuleInterface
from faebryk.core.reference import reference
from faebryk.core.trait import Trait


class Symbol(Module):
    """
    Symbols represent a symbol instance and are bi-directionally
    linked with the module they represent via the `has_linked` trait.
    """

    class Pin(ModuleInterface):
        represents = reference(F.Electrical)

        class has_pin(F.has_reference.decless()):
            """
            Attach to an ElectricalInterface to point back at the pin
            """

            reference: "Symbol.Pin" = reference()

    class TraitT(Trait): ...

    class has_symbol(F.has_reference.decless()):
        """
        Attach to an Module to point back at the pin
        """

        reference: "Symbol" = reference()

    class has_kicad_symbol(TraitT.decless()):
        """
        If a symbol has this trait, then the symbol has a matching KiCAD symbol
        :param symbol_name: The full name of the KiCAD symbol including the library name
        """

        def __init__(self, symbol_name: str):
            super().__init__()
            self.symbol_name = symbol_name

    pins: dict[str, Pin]
    represents = reference(Module)

    @classmethod
    def with_component(cls, component: Module, pin_map: dict[str, F.Electrical]):
        sym = cls()
        sym.represents = component
        component.add(cls.has_symbol(sym))

        sym.pins = {}
        for pin_name, e_pin in pin_map.items():
            pin = cls.Pin()
            pin.represents = e_pin
            e_pin.add(cls.Pin.has_pin(pin))
            sym.pins[pin_name] = pin

        return sym



# This file is part of the faebryk project
# SPDX-License-Identifier: MIT


import csv
import logging
import re
import subprocess

logger = logging.getLogger(__name__)

# Expects a csv file in the format: issue_number,title
# Can be generated with gh issue list and some manual editing

with open("issues.txt", "r") as f:
    reader = csv.DictReader(f)
    rows = list(reader)

issues = {row["issue"]: row["title"] for row in rows}

new_titles = {
    issue: re.sub(r"^\[[^\]]*\][ :]*", "", title) for issue, title in issues.items()
}

for issue, title in issues.items():
    logger.info("{}->{}".format(title, new_titles[issue]))


for issue, title in new_titles.items():
    subprocess.run(["gh", "issue", "edit", issue, "--title", title])



import pytest

from atopile.cli.create import AtoTemplate, FabllTemplate
from atopile.parse import parse_text_as_file
from faebryk.libs.picker.api.picker_lib import client


# Test a variety of component types
@pytest.mark.parametrize(
    "lcsc_id,description",
    [
        (2040, "RP2040"),
        (404010, "STM32H7"),
        (80100, "RC4580IDR"),
    ],
)
def test_ato_create_component(lcsc_id, description):
    components = client.fetch_part_by_lcsc(lcsc_id)

    # Create and populate template
    template = AtoTemplate(name="test_component", base="Module")
    template.add_part(components[0])

    # Generate the output
    output = template.dumps()

    # Basic assertions
    assert f'lcsc_id = "C{lcsc_id}"' in output
    assert "manufacturer =" in output
    assert "mpn =" in output

    tree = parse_text_as_file(output)
    assert tree is not None, "Failed to parse generated template"


@pytest.mark.parametrize(
    "lcsc_id,description",
    [
        (2040, "RP2040"),
        (404010, "STM32H7"),
        (80100, "RC4580IDR"),
    ],
)
def test_fabll_create_component(lcsc_id, description):
    components = client.fetch_part_by_lcsc(lcsc_id)

    if not components:
        pytest.skip(f"No components found for LCSC ID {lcsc_id} ({description})")

    # Create and populate template
    template = FabllTemplate(name="test_component", base="Module")
    template.add_part(components[0])

    # Generate the output
    output = template.dumps()

    # Make sure the code is valid
    compile(output, "<string>", "exec")

    # Basic assertions for expected content
    assert "lcsc_id = L.f_field" in output
    assert "descriptive_properties = L.f_field" in output
    assert "attach_via_pinmap" in output



# This file is part of the faebryk project
# SPDX-License-Identifier: MIT

import logging
from typing import Iterable, Sequence

from faebryk.core.module import Module
from faebryk.core.moduleinterface import ModuleInterface
from faebryk.library.can_specialize import can_specialize

logger = logging.getLogger(__name__)


class can_specialize_defined(can_specialize.impl()):
    def __init__(
        self, specializable_types: Sequence[type["Module"] | type["ModuleInterface"]]
    ):
        super().__init__()
        self._specializable_types = specializable_types

    def get_specializable_types(
        self,
    ) -> Iterable[type["Module"] | type["ModuleInterface"]]:
        return self._specializable_types



# This file is part of the faebryk project
# SPDX-License-Identifier: MIT

import logging
from typing import TYPE_CHECKING

from faebryk.core.module import Module
from faebryk.libs.util import not_none

if TYPE_CHECKING:
    from faebryk.libs.picker.picker import Part

logger = logging.getLogger(__name__)


class has_part_picked(Module.TraitT.decless()):
    def __init__(self, part: "Part | None") -> None:
        """
        None = remove part from netlist
        """
        super().__init__()
        if type(self) is has_part_picked and part is None:
            raise ValueError("Use has_part_picked.remove()")
        self.part = part

    def get_part(self) -> "Part":
        return not_none(self.part)

    def try_get_part(self) -> "Part | None":
        return self.part

    @property
    def removed(self) -> bool:
        return self.part is None



# This file is part of the faebryk project
# SPDX-License-Identifier: MIT

import logging
from abc import ABC, abstractmethod
from collections import defaultdict
from contextlib import contextmanager
from dataclasses import dataclass
from enum import StrEnum
from textwrap import indent
from typing import TYPE_CHECKING, Any, Iterable

from rich.progress import Progress

import faebryk.library._F as F
from atopile.cli.console import error_console
from faebryk.core.cpp import Graph
from faebryk.core.graph import GraphFunctions
from faebryk.core.module import Module
from faebryk.core.moduleinterface import ModuleInterface
from faebryk.core.parameter import (
    ConstrainableExpression,
    Is,
    Parameter,
    ParameterOperatable,
)
from faebryk.core.solver.solver import LOG_PICK_SOLVE, Solver
from faebryk.core.solver.utils import Contradiction, ContradictionByLiteral, get_graphs
from faebryk.libs.sets.sets import P_Set
from faebryk.libs.test.times import Times
from faebryk.libs.util import (
    ConfigFlag,
    EquivalenceClasses,
    KeyErrorAmbiguous,
    KeyErrorNotFound,
    Tree,
    indented_container,
    partition,
    try_or,
    unique,
)

if TYPE_CHECKING:
    from faebryk.libs.picker.api.models import Component
    from faebryk.libs.picker.localpick import PickerOption


NO_PROGRESS_BAR = ConfigFlag("NO_PROGRESS_BAR", default=False)

logger = logging.getLogger(__name__)


class DescriptiveProperties(StrEnum):
    manufacturer = "Manufacturer"
    partno = "Partnumber"
    datasheet = "Datasheet"


class Supplier(ABC):
    @abstractmethod
    def attach(self, module: Module, part: "PickerOption"): ...


@dataclass(frozen=True)
class Part:
    partno: str
    supplier: Supplier


class PickError(Exception):
    def __init__(self, message: str, *modules: Module):
        self.message = message
        self.modules = modules

    def __repr__(self):
        return f"{type(self).__name__}({self.modules}, {self.message})"

    def __str__(self):
        return self.message


class PickErrorNotImplemented(PickError):
    def __init__(self, module: Module):
        message = f"Could not pick part for {module}: Not implemented"
        super().__init__(message, module)


class PickVerificationError(PickError):
    def __init__(self, message: str, *modules: Module):
        message = f"Post-pick verification failed for picked parts:\n{message}"
        super().__init__(message, *modules)


class PickErrorChildren(PickError):
    def __init__(self, module: Module, children: dict[Module, PickError]):
        self.children = children

        message = f"Could not pick parts for children of {module}:\n" + "\n".join(
            f"{m}: caused by {v.message}" for m, v in self.get_all_children().items()
        )
        super().__init__(message, module)

    def get_all_children(self):
        return {
            k: v
            for k, v in self.children.items()
            if not isinstance(v, PickErrorChildren)
        } | {
            module: v2
            for v in self.children.values()
            if isinstance(v, PickErrorChildren)
            for module, v2 in v.get_all_children().items()
        }


class NotCompatibleException(Exception):
    def __init__(
        self,
        module: Module,
        component: "Component",
        param: Parameter | None = None,
        c_range: P_Set | None = None,
    ):
        self.module = module
        self.component = component
        self.param = param
        self.c_range = c_range

        if param is None or c_range is None:
            msg = f"{component.lcsc_display} is not compatible with `{module}`"
        else:
            msg = (
                f"`{param}` ({param.try_get_literal_subset()}) is not "
                f"compatible with {component.lcsc_display} ({c_range})"
            )

        super().__init__(msg)


class does_not_require_picker_check(Parameter.TraitT.decless()):
    pass


# FIXME: remove? uses wrong console
class PickerProgress:
    def __init__(self, tree: Tree[Module]):
        self.tree = tree
        self.progress = Progress(
            disable=bool(NO_PROGRESS_BAR),
            transient=True,
            # This uses the error console to properly interleave with logging
            console=error_console,
        )
        leaves = list(tree.leaves())
        count = len(leaves)

        logger.info(f"Picking parts for {count} leaf modules")
        self.task = self.progress.add_task("Picking", total=count)

    def advance(self, module: Module):
        leaf_count = len(list(self.tree.get_subtree(module).leaves()))
        # module is leaf
        if not leaf_count:
            leaf_count = 1
        self.progress.advance(self.task, leaf_count)

    @contextmanager
    def context(self):
        with self.progress:
            yield self


def get_pick_tree(module: Module | ModuleInterface) -> Tree[Module]:
    if isinstance(module, Module):
        module = module.get_most_special()

    tree = Tree()
    merge_tree = tree

    if module.has_trait(F.has_part_picked):
        return tree

    if module.has_trait(F.is_pickable):
        merge_tree = Tree()
        tree[module] = merge_tree

    for child in module.get_children(
        direct_only=True, types=(Module, ModuleInterface), include_root=False
    ):
        child_tree = get_pick_tree(child)
        merge_tree.update(child_tree)

    return tree


def update_pick_tree(tree: Tree[Module]) -> tuple[Tree[Module], bool]:
    if not tree:
        return tree, False

    filtered_tree = Tree(
        (k, sub[0])
        for k, v in tree.items()
        if not k.has_trait(F.has_part_picked) and not (sub := update_pick_tree(v))[1]
    )
    if not filtered_tree:
        return filtered_tree, True

    return filtered_tree, False


def check_missing_picks(module: Module):
    # - not skip self pick
    # - no parent with part picked
    # - not specialized
    # - no module children
    # - no parent with picker

    missing = module.get_children_modules(
        types=Module,
        direct_only=False,
        include_root=True,
        # not specialized
        most_special=True,
        # leaf == no children
        f_filter=lambda m: not m.get_children_modules(
            types=Module, f_filter=lambda x: not isinstance(x, F.Footprint)
        )
        and not isinstance(m, F.Footprint)
        # no parent with part picked
        and not try_or(
            lambda: m.get_parent_with_trait(F.has_part_picked),
            default=False,
            catch=KeyErrorNotFound,
        )
        # no parent with picker
        and not try_or(
            lambda: try_or(
                lambda: m.get_parent_with_trait(F.is_pickable),
                default=False,
                catch=KeyErrorNotFound,
            ),
            default=True,
            catch=KeyErrorAmbiguous,
        ),
    )

    if missing:
        no_fp, with_fp = map(
            list, partition(lambda m: m.has_trait(F.has_footprint), missing)
        )

        if with_fp:
            logger.warning(f"No pickers for {indented_container(with_fp)}")
        if no_fp:
            logger.warning(
                f"No pickers and no footprint for {indented_container(no_fp)}."
                "\nATTENTION: These modules will not appear in netlist or pcb."
            )


def find_independent_groups(
    modules: Iterable[Module], solver: Solver
) -> list[set[Module]]:
    """
    Find groups of modules that are independent of each other.
    """
    from faebryk.core.solver.defaultsolver import DefaultSolver

    modules = set(modules)

    if (
        not isinstance(solver, DefaultSolver)
        or (state := solver.reusable_state) is None
    ):
        # Find params aliased to lits
        aliased = EquivalenceClasses[ParameterOperatable]()
        lits = dict[ParameterOperatable, ParameterOperatable.Literal]()
        for e in GraphFunctions(*get_graphs(modules)).nodes_of_type(Is):
            if not e.constrained:
                continue
            aliased.add_eq(*e.operatable_operands)
            if e.get_operand_literals():
                lits[e.get_operand_operatables().pop()] = next(
                    iter(e.get_operand_literals().values())
                )
        for alias_group in aliased.get():
            lits_eq = unique((lits[p] for p in alias_group if p in lits), lambda x: x)
            if not lits_eq:
                continue
            if len(lits_eq) > 1:
                # TODO
                raise ContradictionByLiteral("", [], [], None)
            for p in alias_group:
                lits[p] = lits_eq[0]

        # find params related to each other
        param_eqs = EquivalenceClasses[Parameter]()
        for e in GraphFunctions(*get_graphs(modules)).nodes_of_type(
            ConstrainableExpression
        ):
            if not e.constrained:
                continue
            ps = e.get_operand_parameters(recursive=True)
            ps.difference_update(lits.keys())
            param_eqs.add_eq(*ps)

        # find modules that are dependent on each other
        module_eqs = EquivalenceClasses[Module](modules)
        for p_eq in param_eqs.get():
            p_modules = {
                m
                for p in p_eq
                if (parent := p.get_parent()) is not None
                and isinstance(m := parent[0], Module)
                and m in modules
            }
            module_eqs.add_eq(*p_modules)
        out = module_eqs.get()
        logger.debug(indented_container(out, recursive=True))
        return out

    graphs = EquivalenceClasses()
    graph_to_m = defaultdict[Graph, set[Module]](set)
    for m in modules:
        params = m.get_trait(F.is_pickable_by_type).get_parameters().values()
        new_params = {state.data.mutation_map.map_forward(p).maps_to for p in params}
        m_graphs = get_graphs(new_params)
        graphs.add_eq(*m_graphs)
        for g in m_graphs:
            graph_to_m[g].add(m)

    graph_groups: list[set[Graph]] = graphs.get()
    module_groups = [{m for g in gg for m in graph_to_m[g]} for gg in graph_groups]
    return module_groups


def _list_to_hack_tree(modules: Iterable[Module]) -> Tree[Module]:
    return Tree({m: Tree() for m in modules})


def pick_topologically(
    tree: Tree[Module],
    solver: Solver,
    progress: PickerProgress | None = None,
):
    # TODO implement backtracking

    import faebryk.libs.picker.api.picker_lib as picker_lib

    timings = Times(name="pick")
    logger.info(f"Picking {len(tree)} modules")

    explicit_modules = [
        m
        for m in tree.keys()
        if m.has_trait(F.is_pickable_by_part_number)
        or m.has_trait(F.is_pickable_by_supplier_id)
    ]
    logger.info(f"Picking {len(explicit_modules)} explicit parts")
    explicit_parts = picker_lib._find_modules(
        _list_to_hack_tree(explicit_modules), solver
    )
    for m, parts in explicit_parts.items():
        part = parts[0]
        picker_lib.attach_single_no_check(m, part, solver)
    if explicit_parts:
        tree, _ = update_pick_tree(tree)

    tree_backup = set(tree.keys())
    _pick_count = len(tree)

    def _get_candidates(_tree: Tree[Module]):
        # with timings.as_global("pre-solve"):
        #    solver.simplify(*get_graphs(tree.keys()))
        try:
            with timings.as_global("new estimates"):
                # Rerun solver for new system
                solver.update_superset_cache(*_tree)
        except Contradiction as e:
            raise PickError(str(e), *_tree.keys())
        with timings.as_global("get candidates"):
            candidates = picker_lib.get_candidates(_tree, solver)
        if LOG_PICK_SOLVE:
            logger.info(
                "Candidates: \n\t"
                f"{'\n\t'.join(f'{m}: {len(p)}' for m, p in candidates.items())}"
            )
        return candidates

    def _update_progress(done: Iterable[tuple[Module, Any]] | Module):
        if not progress:
            return

        if isinstance(done, Module):
            modules = [done]
        else:
            modules = [m for m, _ in done]

        for m in modules:
            progress.advance(m)

    timings.add("setup")

    # Works by looking for each module again for compatible parts
    # If no compatible parts are found,
    #   it means we need to backtrack or there is no solution
    with timings.as_global("parallel slow-pick", context=True):
        if tree:
            logger.info(f"Picking {len(tree)} modules in parallel")
        while tree:
            try:
                candidates = _get_candidates(tree)
            except Contradiction as e:
                # TODO better error, also remove from get_candidates
                raise PickError(
                    "Contradiction in system."
                    "Unfixable due to lack of backtracking." + str(e),
                    *[],
                )
            groups = find_independent_groups(candidates.keys(), solver)
            # pick module with least candidates first
            picked = [
                (
                    m := min(group, key=lambda _m: len(candidates[_m])),
                    candidates[m][0],
                )
                for group in groups
            ]
            logger.info(
                f"Picking {len(groups)} independent groups: "
                f"{indented_container([m for m, _ in picked])}"
            )
            for m, part in picked:
                picker_lib.attach_single_no_check(m, part, solver)

            _update_progress(picked)
            tree, _ = update_pick_tree(tree)

    if _pick_count:
        logger.info(
            f"Slow-picked parts in {timings.get_formatted('parallel slow-pick')}"
        )

    logger.info(f"Picked complete: picked {_pick_count} parts")
    logger.info("Verify design")
    try:
        solver.update_superset_cache(*tree_backup)
    except Contradiction as e:
        raise PickVerificationError(str(e), *tree_backup) from e


# TODO should be a Picker
def pick_part_recursively(module: Module, solver: Solver):
    pick_tree = get_pick_tree(module)
    check_missing_picks(module)

    try:
        pick_topologically(pick_tree, solver)
    # FIXME: This does not get called anymore
    except PickErrorChildren as e:
        failed_parts = e.get_all_children()
        for m, sube in failed_parts.items():
            logger.error(
                f"Could not find pick for {m}:\n {sube.message}\n"
                f"Params:\n{indent(m.pretty_params(solver), prefix=' '*4)}"
            )
        raise



# This file is part of the faebryk project
# SPDX-License-Identifier: MIT

import itertools
import logging

import faebryk.library._F as F  # noqa: F401
from faebryk.core.module import Module
from faebryk.libs.library import L  # noqa: F401
from faebryk.libs.units import P  # noqa: F401

logger = logging.getLogger(__name__)


class Analog_Devices_ADM2587EBRWZ_ReferenceDesign(Module):
    """
    Reference implementation of ADM2587EBRWZ isolated RS485 transceiver
    """

    # ----------------------------------------
    #     modules, interfaces, parameters
    # ----------------------------------------
    transceiver = L.f_field(F.Analog_Devices_ADM2587EBRWZ)(full_duplex=False)

    # ----------------------------------------
    #                 traits
    # ----------------------------------------
    @L.rt_field
    def pcb_layout(self):
        from faebryk.exporters.pcb.layout.absolute import LayoutAbsolute
        from faebryk.exporters.pcb.layout.typehierarchy import LayoutTypeHierarchy

        Point = F.has_pcb_position.Point
        L = F.has_pcb_position.layer_type
        LVL = LayoutTypeHierarchy.Level

        return F.has_pcb_layout_defined(
            layout=LayoutTypeHierarchy(
                layouts=[
                    LVL(
                        mod_type=F.Analog_Devices_ADM2587EBRWZ,
                        layout=LayoutAbsolute(Point((0, 0, 0, L.NONE))),
                    ),
                ]
            )
        )

    def __preinit__(self):
        # ------------------------------------
        #           connections
        # ------------------------------------
        self.transceiver.power_isolated_in.connect(self.transceiver.power_isolated_out)

        # ------------------------------------
        #          parametrization
        # ------------------------------------
        # decoupling unisolated power
        power_unisolated_capacitors = (
            self.transceiver.power_unisolated.decoupled.decouple(self)
            .specialize(F.MultiCapacitor(4))
            .capacitors
        )
        capacitance_values = [100, 10]  # in nF

        for cap, value in zip(
            power_unisolated_capacitors, itertools.cycle(capacitance_values)
        ):
            cap.capacitance.constrain_subset(
                L.Range.from_center_rel(value * P.nF, 0.05)
            )

        # decoupling isolated power in
        for i, cap in enumerate(
            self.transceiver.power_isolated_in.decoupled.decouple(self)
            .specialize(F.MultiCapacitor(2))
            .capacitors
        ):
            cap.capacitance.constrain_subset(
                L.Range.from_center_rel(capacitance_values[i] * P.nF, 0.05)
            )
        # decoupling isolated power out
        for i, cap in enumerate(
            self.transceiver.power_isolated_out.decoupled.decouple(self)
            .specialize(F.MultiCapacitor(2))
            .capacitors
        ):
            cap.capacitance.constrain_subset(
                L.Range.from_center_rel(capacitance_values[i] * P.nF, 0.05)
            )



# This file is part of the faebryk project
# SPDX-License-Identifier: MIT

from abc import abstractmethod
from enum import Enum, auto
from typing import Self

import faebryk.library._F as F
from faebryk.core.module import Module
from faebryk.libs.library import L


class ElectricLogic(F.ElectricSignal):
    class has_pulls(F.Logic.TraitT):
        @abstractmethod
        def get_pulls(self) -> tuple[F.Resistor | None, F.Resistor | None]: ...

    class has_pulls_defined(has_pulls.impl()):
        def __init__(self, up: F.Resistor | None, down: F.Resistor | None) -> None:
            super().__init__()
            self.up = up
            self.down = down

        def get_pulls(self) -> tuple[F.Resistor | None, F.Resistor | None]:
            return self.up, self.down

    class can_be_pulled(F.Logic.TraitT):
        @abstractmethod
        def pull(self, up: bool, owner: Module) -> F.Resistor: ...

    class can_be_pulled_defined(can_be_pulled.impl()):
        def __init__(self, line: F.Electrical, ref: F.ElectricPower) -> None:
            super().__init__()
            self.ref = ref
            self.line = line

        def pull(self, up: bool, owner: Module):
            obj = self.obj

            up_r, down_r = None, None
            if obj.has_trait(ElectricLogic.has_pulls):
                up_r, down_r = obj.get_trait(ElectricLogic.has_pulls).get_pulls()

            if up and up_r:
                return up_r
            if not up and down_r:
                return down_r

            resistor = F.Resistor()
            name = obj.get_name(accept_no_parent=True)
            # TODO handle collisions
            if up:
                owner.add(resistor, f"pull_up_{name}")
                up_r = resistor
            else:
                owner.add(resistor, f"pull_down_{name}")
                down_r = resistor

            self.line.connect_via(resistor, self.ref.hv if up else self.ref.lv)

            obj.add(ElectricLogic.has_pulls_defined(up_r, down_r))
            return resistor

    # class can_be_buffered(Trait):
    #    @abstractmethod
    #    def buffer(self):
    #        ...
    #
    #
    # class can_be_buffered_defined(can_be_buffered.impl()):
    #    def __init__(self, line: "ElectricLogic") -> None:
    #        super().__init__()
    #        self.line = line
    #
    #    def buffer(self):
    #        obj = self.obj
    #
    #        if hasattr(obj, "buffer"):
    #            return cast_assert(SignalBuffer, getattr(obj, "buffer"))
    #
    #        buffer = SignalBuffer()
    #        obj.buffer = buffer
    #        self.line.connect(buffer.logic_in)
    #
    #        return buffer.logic_out

    class PushPull(Enum):
        PUSH_PULL = auto()
        OPEN_DRAIN = auto()
        OPEN_SOURCE = auto()

    # ----------------------------------------
    #     modules, interfaces, parameters
    # ----------------------------------------
    push_pull = L.p_field(
        domain=L.Domains.ENUM(PushPull),
    )

    # ----------------------------------------
    #                 traits
    # ----------------------------------------
    @L.rt_field
    def pulled(self):
        return ElectricLogic.can_be_pulled_defined(self.line, self.reference)

    specializable_types = L.f_field(F.can_specialize_defined)([F.Logic])

    # ----------------------------------------
    #                functions
    # ----------------------------------------
    def set(self, on: bool):
        """
        Set the logic signal by directly connecting to the reference.
        """
        r = self.reference
        self.line.connect(r.hv if on else r.lv)

    def set_weak(self, on: bool, owner: Module):
        """
        Set the logic signal by connecting to the reference via a pull resistor.
        """
        return self.get_trait(self.can_be_pulled).pull(up=on, owner=owner)

    def connect_shallow(
        self,
        other: Self,
        signal: bool = False,
        reference: bool = False,
        lv: bool = False,
    ) -> Self:
        # TODO this should actually use shallow links
        assert not (signal and reference)
        assert not (lv and reference)

        # TODO make custom LinkDirectShallow that also allows the specified params
        if signal:
            self.line.connect(other.line)
        if reference:
            self.reference.connect(other.reference)
        if lv:
            self.reference.lv.connect(other.reference.lv)

        return super().connect_shallow(other)



from pathlib import Path

import faebryk.library._F as F

HERE = Path(__file__).parent

KNOWN_PACKAGES_TO_FOOTPRINT = {
    # Capacitors
    F.has_package.Package.C0201: HERE / "C_0201_0603Metric.kicad_mod",
    F.has_package.Package.C0402: HERE / "C_0402_1005Metric.kicad_mod",
    F.has_package.Package.C0603: HERE / "C_0603_1608Metric.kicad_mod",
    F.has_package.Package.C0805: HERE / "C_0805_2012Metric.kicad_mod",
    F.has_package.Package.C1206: HERE / "C_1206_3216Metric.kicad_mod",
    F.has_package.Package.C1210: HERE / "C_1210_3225Metric.kicad_mod",
    F.has_package.Package.C1808: HERE / "C_1808_4520Metric.kicad_mod",
    F.has_package.Package.C1812: HERE / "C_1812_4532Metric.kicad_mod",
    F.has_package.Package.C1825: HERE / "C_1825_4564Metric.kicad_mod",
    F.has_package.Package.C2220: HERE / "C_2220_5750Metric.kicad_mod",
    F.has_package.Package.C2225: HERE / "C_2225_5664Metric.kicad_mod",
    F.has_package.Package.C3640: HERE / "C_3640_9110Metric.kicad_mod",
    F.has_package.Package.C01005: HERE / "C_01005_0402Metric.kicad_mod",
    # Resistors
    F.has_package.Package.R0201: HERE / "R_0201_0603Metric.kicad_mod",
    F.has_package.Package.R0402: HERE / "R_0402_1005Metric.kicad_mod",
    F.has_package.Package.R0603: HERE / "R_0603_1608Metric.kicad_mod",
    F.has_package.Package.R0805: HERE / "R_0805_2012Metric.kicad_mod",
    F.has_package.Package.R0815: HERE / "R_0815_2038Metric.kicad_mod",
    F.has_package.Package.R1020: HERE / "R_1020_2550Metric.kicad_mod",
    F.has_package.Package.R1206: HERE / "R_1206_3216Metric.kicad_mod",
    F.has_package.Package.R1210: HERE / "R_1210_3225Metric.kicad_mod",
    F.has_package.Package.R1218: HERE / "R_1218_3246Metric.kicad_mod",
    F.has_package.Package.R1812: HERE / "R_1812_4532Metric.kicad_mod",
    F.has_package.Package.R2010: HERE / "R_2010_5025Metric.kicad_mod",
    F.has_package.Package.R2512: HERE / "R_2512_6332Metric.kicad_mod",
    F.has_package.Package.R2816: HERE / "R_2816_7142Metric.kicad_mod",
    F.has_package.Package.R4020: HERE / "R_4020_10251Metric.kicad_mod",
    F.has_package.Package.R01005: HERE / "R_01005_0402Metric.kicad_mod",
    # Inductors
    F.has_package.Package.L0201: HERE / "L_0201_0603Metric.kicad_mod",
    F.has_package.Package.L0402: HERE / "L_0402_1005Metric.kicad_mod",
    F.has_package.Package.L0603: HERE / "L_0603_1608Metric.kicad_mod",
    F.has_package.Package.L0805: HERE / "L_0805_2012Metric.kicad_mod",
    F.has_package.Package.L1008: HERE / "L_1008_2520Metric.kicad_mod",
    F.has_package.Package.L1206: HERE / "L_1206_3216Metric.kicad_mod",
    F.has_package.Package.L1210: HERE / "L_1210_3225Metric.kicad_mod",
    F.has_package.Package.L1806: HERE / "L_1806_4516Metric.kicad_mod",
    F.has_package.Package.L1812: HERE / "L_1812_4532Metric.kicad_mod",
    F.has_package.Package.L2010: HERE / "L_2010_5025Metric.kicad_mod",
    F.has_package.Package.L2512: HERE / "L_2512_6332Metric.kicad_mod",
    F.has_package.Package.L01005: HERE / "L_01005_0402Metric.kicad_mod",
}



from pathlib import Path
from textwrap import dedent

from atopile.config import PROJECT_CONFIG_FILENAME, Dependency, config

TEST_CONFIG_TEXT = dedent(
    """
    ato-version: ^0.2.0
    builds:
      debug:
        entry: elec/src/debug.ato:Debug
    # comments
    dependencies:
    - tps63020dsjr # comments
    - usb-connectors ^v2.0.1
    - esp32-s3
    - name: rp2040
      path: ../rp2040
      version_spec: ^v0.0.1
      link_broken: true
    """
).lstrip()


def test_roundtrip(tmp_path: Path):
    config_path = tmp_path / PROJECT_CONFIG_FILENAME
    config_path.write_text(TEST_CONFIG_TEXT)
    config.project_dir = tmp_path

    assert config.project.ato_version == "^0.2.0"
    assert config.project.dependencies is not None
    assert config.project.dependencies[0].name == "tps63020dsjr"
    assert config.project.dependencies[1].name == "usb-connectors"
    assert config.project.dependencies[1].version_spec == "^v2.0.1"
    assert config.project.dependencies[2].name == "esp32-s3"
    assert config.project.dependencies[3].name == "rp2040"
    assert config.project.dependencies[3].path == Path("../rp2040")
    assert config.project.dependencies[3].version_spec == "^v0.0.1"
    assert config.project.dependencies[3].link_broken is True

    config.update_project_config(lambda data, new_data: data, {})

    assert config_path.read_text() == TEST_CONFIG_TEXT


def test_update_project_config(tmp_path: Path):
    config_path = tmp_path / PROJECT_CONFIG_FILENAME
    config_path.write_text(TEST_CONFIG_TEXT)
    config.project_dir = tmp_path

    # Make some changes and check that they are reflected in the config
    dep1 = Dependency(name="usb-connectors", version_spec="^v0.0.1", path=Path("test"))
    dep2 = Dependency(name="esp32-s3", version_spec="^v0.0.1", path=Path("../esp32-s3"))

    def add_dependency(config_data, new_data):
        config_data["dependencies"] = config_data["dependencies"] + [new_data]
        return config_data

    config.update_project_config(add_dependency, dep1.model_dump())
    config.update_project_config(add_dependency, dep2.model_dump())

    assert config.project.dependencies is not None

    assert config.project.dependencies[4].path == Path("test")
    assert config.project.dependencies[5].path == Path("../esp32-s3")



# This file is part of the faebryk project
# SPDX-License-Identifier: MIT

import faebryk.library._F as F


class has_kicad_footprint_equal_ifs_defined(F.has_kicad_footprint_equal_ifs):
    def __init__(self, str) -> None:
        super().__init__()
        self.str = str

    def get_kicad_footprint(self):
        return self.str



# This file is part of the faebryk project
# SPDX-License-Identifier: MIT

import logging
import math
from dataclasses import dataclass
from enum import IntEnum

import faebryk.library._F as F
from faebryk.core.module import Module
from faebryk.core.node import Node
from faebryk.core.trait import TraitNotFound
from faebryk.exporters.pcb.kicad.transformer import PCB_Transformer
from faebryk.exporters.pcb.layout.layout import Layout
from faebryk.libs.kicad.fileformats import C_kicad_pcb_file, C_wh
from faebryk.libs.util import KeyErrorNotFound, find, not_none

logger = logging.getLogger(__name__)


KFootprint = C_kicad_pcb_file.C_kicad_pcb.C_pcb_footprint
KPad = KFootprint.C_pad

# TODO move all those helpers and make them more general and precise


@dataclass
class Params:
    distance_between_pad_edges: float = 1
    extra_rotation_of_footprint: float = 0


class Side(IntEnum):
    Right = 0
    Bottom = 90
    Left = 180
    Top = 270

    def rot(self, angle=90):
        return type(self)((self + angle) % 360)

    def rot_vector(self, vec: tuple[float, float]):
        x, y = vec
        if self == Side.Right:
            return x, y
        elif self == Side.Bottom:
            return -y, x
        elif self == Side.Left:
            return -x, -y
        elif self == Side.Top:
            return y, -x
        else:
            assert False


def _get_pad_side(fp: KFootprint, pad: KPad) -> Side:
    # TODO this is just a heuristic
    # we better need to check for the bbox

    # relative to fp center
    hyst = 1.1
    right_nleft = Side.Right if pad.at.x > 0 else Side.Left
    bottom_ntop = Side.Bottom if pad.at.y > 0 else Side.Top
    if abs(pad.at.x) > abs(pad.at.y) * hyst:
        pos_side = [right_nleft]
    elif abs(pad.at.y) > abs(pad.at.x) * hyst:
        pos_side = [bottom_ntop]
    else:
        # corner
        if abs(pad.at.x) > abs(pad.at.y):
            pos_side = [right_nleft, bottom_ntop]
        else:
            pos_side = [bottom_ntop, right_nleft]

    # pad size as heuristic
    rot = (fp.at.r - pad.at.r) % 360
    assert rot in (0, 90, 180, 270)
    assert pad.size.h is not None
    if (pad.size.h > pad.size.w) == (rot in (0, 180)):
        pos_rot = {Side.Top, Side.Bottom}
    else:
        pos_rot = {Side.Right, Side.Left}

    match = set(pos_side) & pos_rot
    if match:
        assert len(match) == 1
        return next(iter(match))

    logger.warning(
        f"Uncertain about pad position of `{fp.name}"
        f"|{fp.propertys['Reference'].value}.{pad.name}`: "
        f"Heuristic at: `{pos_side}`, "
        f"Heuristic size: `{pos_rot} ||| {rot=} {pad.at=} {pad.size=}`",
        extra={"markdown": True},
    )
    return pos_side[0]


type V2D = tuple[float, float]


def _vec_pad_center_to_edge(size: C_wh, side: Side):
    assert size.h is not None
    if side == Side.Top:
        return 0, 0 - size.h / 2
    elif side == Side.Bottom:
        return 0, 0 + size.h / 2
    elif side == Side.Left:
        return 0 - size.w / 2, 0
    else:
        return 0 + size.w / 2, 0


def _next_to_pad(
    fp: KFootprint, spad: KPad, dfp: KFootprint, dpad: KPad, params: Params
):
    # TODO determine distance based on pads & footprint size
    distance = params.distance_between_pad_edges
    dfp_bbox = PCB_Transformer.get_footprint_silkscreen_bbox(dfp)
    assert dfp_bbox is not None
    dfp_w = dfp_bbox[1][0] - dfp_bbox[0][0]

    # heuristic
    # TODO use pad pitch instead
    assert spad.size.h
    assert dpad.size.h
    hs = spad.size.h
    hd = dpad.size.h
    # TODO rotation
    if hs < hd:
        no = fp.pads.index(spad)
        tol = 1.1
        extra_distance = dfp_w * tol
        steps = int(math.ceil(hd / hs))
        distance += (no % steps) * extra_distance

    def _add(v1: V2D, v2: V2D) -> V2D:
        return v1[0] + v2[0], v1[1] + v2[1]

    def _sub(v1: V2D, v2: V2D) -> V2D:
        return v1[0] - v2[0], v1[1] - v2[1]

    side = _get_pad_side(fp, spad)

    # rotate fp to let pads face each other (+extra rotation)
    dside = _get_pad_side(dfp, dpad)
    fp_rot_rel_to_source = (
        side.rot() - dside.rot() - 180 + params.extra_rotation_of_footprint
    ) % 360

    vec_distance_directed = side.rot_vector((distance, 0))
    vec_distance_from_pad_center = _add(vec_distance_directed, (spad.at.x, spad.at.y))

    # make vec distance between edges instead of center
    _spad_edge_vec = _vec_pad_center_to_edge(spad.size, side)
    _dpad_edge_vec = Side(fp_rot_rel_to_source).rot_vector(
        _vec_pad_center_to_edge(dpad.size, dside)
    )
    vec_distance_from_pad_edge = _sub(
        _add(vec_distance_from_pad_center, _spad_edge_vec), _dpad_edge_vec
    )

    # get vector to fp center (instead of pad center)
    _vec_pad_to_fp = Side(fp_rot_rel_to_source).rot_vector((-dpad.at.x, -dpad.at.y))
    vec_to_fp_center = _add(vec_distance_from_pad_edge, _vec_pad_to_fp)

    if logger.isEnabledFor(logging.DEBUG):
        logger.debug(
            f"Next to pad: {fp.name}|{fp.propertys['Reference'].value}.{spad.name}"
            f" with {dfp.name}|{dfp.propertys['Reference'].value}.{dpad.name}:"
            f"\n     {fp.at=}"
            f"\n     {spad.at=} | {spad.size=}"
            f"\n     {dpad.at=} | {dpad.size=}"
            f"\n     {side=}"
            f"\n     {dside=}"
            f"\n     {fp_rot_rel_to_source=}"
            f"\n     {vec_distance_directed=}"
            f"\n     {vec_distance_from_pad_center=}"
            f"\n     {_spad_edge_vec=} | {_dpad_edge_vec=}"
            f"\n     {vec_distance_from_pad_edge=}"
            f"\n     {vec_to_fp_center=}"
        )

    # in fp system clockwise, in pcb counter clockwise
    rot_pcb_coord_system = (-fp_rot_rel_to_source) % 360

    return (*vec_to_fp_center, rot_pcb_coord_system)


def place_next_to_pad(
    module: Module,
    pad: F.Pad,
    params: Params,
):
    kfp, kpad = pad.get_trait(PCB_Transformer.has_linked_kicad_pad).get_pad()
    if len(kpad) != 1:
        raise NotImplementedError()
    kpad = kpad[0]

    nfp = module.get_trait(F.has_footprint).get_footprint()
    npad = find(
        nfp.get_children(direct_only=True, types=F.Pad),
        lambda p: p.net.is_connected_to(pad.net),
    )
    nkfp, nkpad = npad.get_trait(PCB_Transformer.has_linked_kicad_pad).get_pad()
    if len(nkpad) != 1:
        raise NotImplementedError()
    nkpad = nkpad[0]

    pos = _next_to_pad(kfp, kpad, nkfp, nkpad, params)

    module.add(
        F.has_pcb_position_defined_relative_to_parent(
            (
                *pos,
                F.has_pcb_position.layer_type.NONE,
            )
        )
    )


def place_next_to(
    parent_intf: F.Electrical,
    module: Module,
    params: Params,
    route: bool = False,
):
    try:
        pads_intf = parent_intf.get_trait(F.has_linked_pad).get_pads()
    except TraitNotFound:
        logger.warning(
            f"No linked pads found for interface {parent_intf}."
            " Make sure the footprint is properly attached."
            " And that the component is attached "
            "to an interface of a module with a footprint."
        )
        return

    if len(pads_intf) == 0:
        raise KeyErrorNotFound()

    children = module.get_children_modules(
        direct_only=False,
        types=Module,
        f_filter=lambda m: m.has_trait(F.has_footprint)
        and not m.has_trait(F.has_pcb_position),
        include_root=True,
    )

    for parent_pad, child in zip(pads_intf, children):
        intf = find(
            child.get_children(direct_only=True, types=F.Electrical),
            lambda x: x.is_connected_to(parent_intf),
        )

        logger.debug(f"Placing {intf} next to {parent_pad}")
        place_next_to_pad(child, parent_pad, params)

        if route:
            intf.add(
                F.has_pcb_routing_strategy_greedy_direct_line(extra_pads=[parent_pad])
            )


class LayoutHeuristicElectricalClosenessDecouplingCaps(Layout):
    Parameters = Params

    def __init__(self, params: Params | None = None):
        super().__init__()
        self._params = params or Params()

    def apply(self, *node: Node):
        # Remove nodes that have a position defined
        node = tuple(n for n in node if not n.has_trait(F.has_pcb_position))

        for n in node:
            assert isinstance(n, F.Capacitor)
            power = not_none(n.get_parent_of_type(F.ElectricPower))

            place_next_to(power.hv, n, route=True, params=self._params)

    @staticmethod
    def find_module_candidates(node: Node):
        return Module.get_children_modules(
            node,
            direct_only=False,
            types=F.Capacitor,
            f_filter=lambda c: c.get_parent_of_type(
                F.ElectricPower, direct_only=True, include_root=False
            )
            is not None,
        )

    @classmethod
    def add_to_all_suitable_modules(cls, node: Node, params: Params | None = None):
        layout = cls(params)
        candidates = cls.find_module_candidates(node)
        for c in candidates:
            logger.debug(f"Adding {cls.__name__} to {c}")
            c.add(F.has_pcb_layout_defined(layout))
        return candidates



# This file is part of the faebryk project
# SPDX-License-Identifier: MIT

import faebryk.library._F as F
from faebryk.core.module import Module
from faebryk.libs.library import L
from faebryk.libs.units import P


class SNx4LVC541A(Module):
    """
    The SN54LVC541A octal buffer/driver is designed for
    3.3-V VCC) 2.7-V to 3.6-V VCC operation, and the SN74LVC541A
    octal buffer/driver is designed for 1.65-V to 3.6-V VCC operation.
    """

    # ----------------------------------------
    #     modules, interfaces, parameters
    # ----------------------------------------
    A = L.list_field(8, F.ElectricLogic)
    Y = L.list_field(8, F.ElectricLogic)

    power: F.ElectricPower

    OE = L.list_field(2, F.ElectricLogic)

    # ----------------------------------------
    #                traits
    # ----------------------------------------
    designator_prefix = L.f_field(F.has_designator_prefix)(
        F.has_designator_prefix.Prefix.U
    )
    datasheet = L.f_field(F.has_datasheet_defined)(
        "https://www.ti.com/lit/ds/symlink/sn74lvc541a.pdf?ts=1718881644774&ref_url=https%253A%252F%252Fwww.mouser.ie%252F"
    )

    def __preinit__(self):
        # ----------------------------------------
        #                parameters
        # ----------------------------------------
        self.power.voltage.constrain_le(3.6 * P.V)

        # ----------------------------------------
        #                aliases
        # ----------------------------------------

        # ----------------------------------------
        #                connections
        # ----------------------------------------
        # FIXME
        # self.power.decoupled.decouple()

    @L.rt_field
    def single_electric_reference(self):
        return F.has_single_electric_reference_defined(
            F.ElectricLogic.connect_all_module_references(self)
        )



# This file is part of the faebryk project
# SPDX-License-Identifier: MIT

from faebryk.core.module import Module


class has_descriptive_properties(Module.TraitT):
    def get_properties(self) -> dict[str, str]:
        raise NotImplementedError()

    def add_properties(self, properties: dict[str, str]):
        raise NotImplementedError()



# This file is part of the faebryk project
# SPDX-License-Identifier: MIT

import faebryk.library._F as F
from faebryk.core.moduleinterface import ModuleInterface
from faebryk.libs.library import L


class SWD(ModuleInterface):
    clk: F.ElectricLogic
    dio: F.ElectricLogic
    swo: F.ElectricLogic
    reset: F.ElectricLogic

    @L.rt_field
    def single_electric_reference(self):
        return F.has_single_electric_reference_defined(
            F.ElectricLogic.connect_all_module_references(self)
        )



# This file is part of the faebryk project
# SPDX-License-Identifier: MIT

import logging
from enum import StrEnum
from pathlib import Path

from faebryk.core.graph import GraphFunctions
from faebryk.core.module import Module
from faebryk.core.parameter import EnumDomain, EnumSet, Parameter
from faebryk.core.solver.solver import Solver
from faebryk.libs.util import KeyErrorAmbiguous, KeyErrorNotFound, once

logger = logging.getLogger(__name__)


class has_package(Module.TraitT.decless()):
    """Module has a package parameter"""

    class Package(StrEnum):
        # Capacitors
        C0201 = "C0201"
        C0402 = "C0402"
        C0603 = "C0603"
        C0805 = "C0805"
        C1206 = "C1206"
        C1210 = "C1210"
        C1808 = "C1808"
        C1812 = "C1812"
        C1825 = "C1825"
        C2220 = "C2220"
        C2225 = "C2225"
        C3640 = "C3640"
        C01005 = "C01005"

        # Resistors
        R0201 = "R0201"
        R0402 = "R0402"
        R0603 = "R0603"
        R0805 = "R0805"
        R0815 = "R0815"
        R1020 = "R1020"
        R1206 = "R1206"
        R1210 = "R1210"
        R1218 = "R1218"
        R1812 = "R1812"
        R2010 = "R2010"
        R2512 = "R2512"
        R2816 = "R2816"
        R4020 = "R4020"
        R01005 = "R01005"

        # Inductors
        L0201 = "L0201"
        L0402 = "L0402"
        L0603 = "L0603"
        L0805 = "L0805"
        L1008 = "L1008"
        L1206 = "L1206"
        L1210 = "L1210"
        L1806 = "L1806"
        L1812 = "L1812"
        L2010 = "L2010"
        L2512 = "L2512"
        L01005 = "L01005"

    def __init__(self, *package_candidates: str | Package) -> None:
        super().__init__()
        # Saved because it makes validation way easier
        self._enum_set = {self.Package(c) for c in package_candidates}
        self.package = Parameter(domain=EnumDomain(self.Package))
        self.package.constrain_subset(EnumSet(*self._enum_set))

        # Cache the resolved package
        self._package: "has_package.Package | None" = None

    def get_package(self, solver: Solver) -> Package:
        if self._package is not None:
            return self._package

        package_superset = solver.inspect_get_known_supersets(self.package)

        if package_superset.is_empty():
            raise KeyErrorNotFound()

        if not package_superset.is_single_element():
            # TODO: add the actual candidates to the exception
            raise KeyErrorAmbiguous([])

        # We have guaranteed `.any()` returns only one thing
        # package_superset.any() is a similar enum to a package, however it's not
        # actually a member of the has_package.Package enum.
        self._package = self.Package(package_superset.any().value)

        return self._package

    def try_get_package(self, solver: Solver | None = None) -> Package | None:
        if solver is None:
            return self._package

        try:
            return self.get_package(solver)
        except (KeyErrorNotFound, KeyErrorAmbiguous):
            return None

    @classmethod
    def standardize_footprints(cls, app: Module, solver: Solver) -> None:
        """
        Attach standard footprints for known packages

        This must be done before the create_footprint_library is run
        """
        import faebryk.library._F as F
        from atopile.packages import KNOWN_PACKAGES_TO_FOOTPRINT
        from faebryk.libs.kicad.fileformats import C_kicad_footprint_file

        gf = GraphFunctions(app.get_graph())

        # TODO: make this caching global. Shit takes time
        @once
        def _get_footprint(fp_path: Path) -> C_kicad_footprint_file:
            return C_kicad_footprint_file.loads(fp_path)

        for node, pkg_t in gf.nodes_with_trait(cls):
            package = pkg_t.try_get_package(solver)
            if package is None:
                continue

            # Skip nodes with footprints already
            # TODO: consider elevating this to an exception
            if node.has_trait(F.has_footprint):
                continue

            if fp_path := KNOWN_PACKAGES_TO_FOOTPRINT.get(package):
                if can_attach_t := node.try_get_trait(F.can_attach_to_footprint):
                    fp = _get_footprint(fp_path)
                    kicad_fp = F.KicadFootprint.from_file(fp)
                    kicad_fp.add(F.KicadFootprint.has_file(fp_path))
                    can_attach_t.attach(kicad_fp)
                else:
                    # TODO: consider elevating this to an exception
                    logger.warning(
                        "%s has a package requirement but no can_attach_to_footprint"
                        " trait",
                        node,
                    )



# This file is part of the faebryk project
# SPDX-License-Identifier: MIT

import logging

import faebryk.library._F as F
from faebryk.core.module import Module
from faebryk.libs.library import L
from faebryk.libs.picker.picker import DescriptiveProperties
from faebryk.libs.units import P

logger = logging.getLogger(__name__)


# TODO remove generic stuff into EEPROM/i2c device etc
class _M24C08_FMN6TP(Module):
    power: F.ElectricPower
    data: F.I2C
    write_protect: F.ElectricLogic
    address_pin = L.list_field(3, F.ElectricLogic)

    @L.rt_field
    def attach_to_footprint(self):
        x = self
        return F.can_attach_to_footprint_via_pinmap(
            {
                "1": x.address_pin[0].line,
                "2": x.address_pin[1].line,
                "3": x.address_pin[2].line,
                "4": x.power.lv,
                "5": x.data.sda.line,
                "6": x.data.scl.line,
                "7": x.write_protect.line,
                "8": x.power.hv,
            }
        )

    mfn_pn = L.f_field(F.has_descriptive_properties_defined)(
        {
            DescriptiveProperties.manufacturer: "STMicroelectronics",
            DescriptiveProperties.partno: "M24C08-FMN6TP",
        }
    )

    designator_prefix = L.f_field(F.has_designator_prefix)(
        F.has_designator_prefix.Prefix.U
    )

    datasheet = L.f_field(F.has_datasheet_defined)(
        "https://eu.mouser.com/datasheet/2/389/m24c08_r-1849629.pdf"
    )

    @L.rt_field
    def reference_defined(self):
        return F.has_single_electric_reference_defined(
            F.ElectricLogic.connect_all_module_references(self)
        )

    def __preinit__(self):
        self.attach_to_footprint.attach(
            F.SOIC(8, size_xy=(3.9 * P.mm, 4.9 * P.mm), pitch=1.27 * P.mm)
        )

        self.power.voltage.constrain_subset(L.Range(1.7 * P.V, 5.5 * P.V))

    def enable_write_protection(self, owner: Module, protect=True):
        if protect:
            self.write_protect.get_trait(F.ElectricLogic.can_be_pulled).pull(
                up=True, owner=owner
            )
            return
        self.write_protect.get_trait(F.ElectricLogic.can_be_pulled).pull(
            up=False, owner=owner
        )

    def set_address(self, addr: int):
        assert addr < (1 << len(self.address_pin))

        for i, e in enumerate(self.address_pin):
            e.set(addr & (1 << i) != 0)


# TODO rename to ReferenceDesign
class M24C08_FMN6TP(Module):
    ic: _M24C08_FMN6TP

    def __preinit__(self):
        self.ic.data.terminate(owner=self)
        self.ic.power.decoupled.decouple(owner=self).capacitance.constrain_subset(
            L.Range(10 * P.nF * 0.9, 100 * P.nF * 1.1)
        )



# This file is part of the faebryk project
# SPDX-License-Identifier: MIT

import logging
from dataclasses import dataclass

import faebryk.library._F as F
from faebryk.core.parameter import (
    Boolean,
    EnumDomain,
    Numbers,
    Parameter,
    ParameterOperableHasNoLiteral,
)
from faebryk.libs.sets.quantity_sets import (
    Quantity_Interval_Disjoint,
)
from faebryk.libs.sets.sets import BoolSet, EnumSet
from faebryk.libs.units import Quantity, Unit, to_si_str
from faebryk.libs.util import join_if_non_empty

logger = logging.getLogger(__name__)


class has_simple_value_representation_based_on_params_chain(
    F.has_simple_value_representation.impl()
):
    @dataclass
    class Spec:
        param: Parameter
        unit: Unit | Quantity | None = None
        """
        Override unit of param \n
        E.g use VA instead of W
        """
        tolerance: bool = False
        prefix: str = ""
        suffix: str = ""
        default: str | None = ""
        """
        Set to None to require the parameter to be set.
        """

        def _get_value(self) -> str:
            try:
                value = self.param.get_literal()
            except ParameterOperableHasNoLiteral:
                if self.default is not None:
                    return self.default
                raise

            domain = self.param.domain

            # TODO this is probably not the only place we will ever need this big switch
            # consider moving it somewhere else
            if isinstance(domain, EnumDomain):
                if self.tolerance:
                    raise ValueError("tolerance not supported for enum")
                # TODO handle units
                enum = EnumSet.from_value(value)
                if not enum.is_single_element():
                    raise NotImplementedError()
                val = next(iter(enum.elements))
                # TODO not sure I like this
                if isinstance(val.value, str):
                    return val.value
                return val.name

            if isinstance(domain, Boolean):
                if self.tolerance:
                    raise ValueError("tolerance not supported for boolean")
                bool_val = BoolSet.from_value(value)
                if not bool_val.is_single_element():
                    raise NotImplementedError()
                return str(next(iter(bool_val.elements))).lower()

            if isinstance(domain, Numbers):
                unit = self.unit if self.unit is not None else self.param.units
                # TODO If tolerance, maybe hint that it's weird there isn't any
                value_lit = Quantity_Interval_Disjoint.from_value(value)
                if value_lit.is_single_element():
                    return to_si_str(value_lit.min_elem, unit, 2)
                if len(value_lit._intervals.intervals) > 1:
                    raise NotImplementedError()
                center, tolerance = value_lit.as_gapless().as_center_tuple(
                    relative=True
                )
                center_str = to_si_str(center, unit, 2)
                assert isinstance(tolerance, Quantity)
                if self.tolerance and tolerance > 0:
                    tolerance_str = f" ±{to_si_str(tolerance, "%", 0)}"
                    return f"{center_str}{tolerance_str}"
                return center_str

            raise NotImplementedError(f"No support for {domain}")

        def get_value(self) -> str:
            try:
                value = self._get_value()
            except Exception as e:
                if self.default is None:
                    raise
                logger.debug(f"Failed to get value for `{self.param}`: {e}")
                return ""
            return join_if_non_empty(
                " ",
                self.prefix,
                value,
                self.suffix,
            )

    def __init__(self, *specs: Spec, prefix: str = "", suffix: str = "") -> None:
        super().__init__()
        self.specs = specs
        self.prefix = prefix
        self.suffix = suffix

    def get_value(self) -> str:
        return join_if_non_empty(
            " ",
            self.prefix,
            *[s.get_value() for s in self.specs],
            self.suffix,
        )



# This file is part of the faebryk project
# SPDX-License-Identifier: MIT

import faebryk.library._F as F
from faebryk.core.moduleinterface import ModuleInterface


class UART(ModuleInterface):
    base_uart: F.UART_Base
    rts: F.ElectricLogic
    cts: F.ElectricLogic
    dtr: F.ElectricLogic
    dsr: F.ElectricLogic
    dcd: F.ElectricLogic
    ri: F.ElectricLogic

    # TODO: this creates too many connections in some projects
    # @L.rt_field
    # def single_electric_reference(self):
    #    return F.has_single_electric_reference_defined(
    #       F.ElectricLogic.connect_all_module_references(self)
    #   )



# This file is part of the faebryk project
# SPDX-License-Identifier: MIT

from faebryk.core.trait import Trait


class has_kicad_ref(Trait):
    def get_ref(self) -> str:
        raise NotImplementedError()



# This file is part of the faebryk project
# SPDX-License-Identifier: MIT

import logging
import os
import unittest

import pytest

logger = logging.getLogger(__name__)


@pytest.mark.skip(reason="Not implemented")
class TestNetlistKicad2FaebrykProject(unittest.TestCase):
    def test_eq(self):
        self.assertTrue(False)

        from faebryk.exporters.project.faebryk.project_faebryk import from_t1_netlist
        from faebryk.importers.netlist.kicad.netlist_kicad import (
            to_faebryk_t1_netlist,
            to_faebryk_t2_netlist,
        )

        with open(
            os.path.join(
                os.path.dirname(__file__), "../../../common/resources/test.net"
            ),
            "r",
        ) as f:
            test_netlist = f.read()

        t2 = to_faebryk_t2_netlist(test_netlist)
        t1 = to_faebryk_t1_netlist(t2)
        from_t1_netlist(t1)

        # print(prj)

        # from pathlib import Path
        # path = Path("./build/faebryk_prj.py")
        # logger.info("Writing Experiment prj to {}".format(path.absolute()))
        # path.write_text(prj)



# This file is part of the faebryk project
# SPDX-License-Identifier: MIT

import logging

import faebryk.library._F as F
from faebryk.core.module import Module
from faebryk.libs.library import L
from faebryk.libs.picker.picker import DescriptiveProperties
from faebryk.libs.units import P

logger = logging.getLogger(__name__)


class _ESP32_C3_MINI_1(Module):
    """ESP32-C3-MINI-1 module"""

    esp32_c3: F.ESP32_C3
    # TODO: add components as described in the datasheet

    rf_output: F.Electrical
    chip_enable: F.ElectricLogic
    gpio = L.list_field(
        22, F.ElectricLogic
    )  # TODO: Only GPIO 0 to 10 and 18, 19 are exposed
    uart: F.UART_Base
    vdd3v3: F.ElectricPower

    # TODO: connect all components (nodes)

    @L.rt_field
    def single_electric_reference(self):
        return F.has_single_electric_reference_defined(
            F.ElectricLogic.connect_all_module_references(self)
        )

    def __preinit__(self):
        e = self.esp32_c3
        for v33 in (e.vdd3p3, e.vdd3p3_cpu, e.vdd3p3_rtc, e.vdda):
            self.vdd3v3.connect(v33)

        for lhs, rhs in zip(self.gpio, self.esp32_c3.gpio):
            lhs.connect(rhs)

        # TODO: set the following in the pinmux
        # UART0 gpio 20/21

        self.chip_enable.connect(e.enable)

    @L.rt_field
    def attach_to_footprint(self):
        gnd = self.vdd3v3.lv
        self.pinmap_default = {
            "1": gnd,
            "2": gnd,
            "3": self.vdd3v3.hv,
            # 4 is not connected
            "5": self.gpio[2].line,
            "6": self.gpio[3].line,
            # 7 is not connected
            "8": self.chip_enable.line,
            # 9 is not connected
            # 10 is not connected
            "11": gnd,
            "12": self.gpio[0].line,
            "13": self.gpio[1].line,
            "14": gnd,
            # 15 is not connected
            "16": self.gpio[10].line,
            # 17 is not connected
            "18": self.gpio[4].line,
            "19": self.gpio[5].line,
            "20": self.gpio[6].line,
            "21": self.gpio[7].line,
            "22": self.gpio[8].line,
            "23": self.gpio[9].line,
            # 24 is not connected
            # 25 is not connected
            "26": self.gpio[18].line,
            "27": self.gpio[19].line,
            # 28 is not connected
            # 29 is not connected
            "30": self.gpio[20].line,  # uart.rx,
            "31": self.gpio[21].line,  # uart.tx,
            # 32 is not connected
            # 33 is not connected
            # 34 is not connected
            # 35 is not connected
            "36": gnd,
            "37": gnd,
            "38": gnd,
            "39": gnd,
            "40": gnd,
            "41": gnd,
            "42": gnd,
            "43": gnd,
            "44": gnd,
            "45": gnd,
            "46": gnd,
            "47": gnd,
            "48": gnd,
            "49": gnd,
            "50": gnd,
            "51": gnd,
            "52": gnd,
            "53": gnd,
            "54": gnd,
            "55": gnd,
            "56": gnd,
            "57": gnd,
            "58": gnd,
            "59": gnd,
            "60": gnd,
            "61": gnd,
        }
        self.pinmap = dict(self.pinmap_default)

        return F.can_attach_to_footprint_via_pinmap(self.pinmap)

    descriptive_properties = L.f_field(F.has_descriptive_properties_defined)(
        {
            DescriptiveProperties.manufacturer: "Espressif Systems",
            DescriptiveProperties.partno: "ESP32-C3-MINI-1U-H4",
        }
    )

    designator_prefix = L.f_field(F.has_designator_prefix)(
        F.has_designator_prefix.Prefix.U
    )
    datasheet = L.f_field(F.has_datasheet_defined)(
        "https://www.espressif.com/sites/default/files/documentation/esp32-c3-mini-1_datasheet_en.pdf"
    )


# TODO rename to ReferenceDesign
class ESP32_C3_MINI_1(Module):
    ic: _ESP32_C3_MINI_1

    def __preinit__(self):
        self.ic.chip_enable.pulled.pull(up=True, owner=self)

        # connect power decoupling caps
        self.ic.vdd3v3.decoupled.decouple(owner=self).capacitance.constrain_subset(
            L.Range(100 * P.nF, 10 * P.uF)
        )
